# from pyiceberg.typedef import EMPTY_DICT, Identifier, Properties
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    task_resource_options_provider,
)
import ray
import functools
from deltacat.compute.converter.utils.convert_task_options import (
    convert_resource_options_provider,
)
import logging
from deltacat import logs
from deltacat.compute.converter.model.converter_session_params import (
    ConverterSessionParams,
)


from deltacat.compute.converter.constants import DEFAULT_MAX_PARALLEL_DATA_FILE_DOWNLOAD
from deltacat.compute.converter.steps.convert import convert
from deltacat.compute.converter.model.convert_input import ConvertInput
from deltacat.compute.converter.pyiceberg.overrides import (
    fetch_all_bucket_files,
    parquet_files_dict_to_iceberg_data_files,
)
from deltacat.compute.converter.utils.converter_session_utils import (
    construct_iceberg_table_prefix,
)
from deltacat.compute.converter.pyiceberg.update_snapshot_overrides import (
    commit_replace_snapshot,
    commit_append_snapshot,
)
from deltacat.compute.converter.pyiceberg.catalog import load_table
from deltacat.compute.converter.utils.converter_session_utils import (
    group_all_files_to_each_bucket,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def converter_session(params: ConverterSessionParams, **kwargs):
    """
    Convert equality delete to position delete.
    Compute and memory heavy work from downloading equality delete table and compute position deletes
    will be executed on Ray remote tasks.
    """

    catalog = params.catalog
    table_name = params.iceberg_table_name
    iceberg_table = load_table(catalog, table_name)
    enforce_primary_key_uniqueness = params.enforce_primary_key_uniqueness
    data_file_dict, equality_delete_dict, pos_delete_dict = fetch_all_bucket_files(
        iceberg_table
    )
    convert_input_files_for_all_buckets = group_all_files_to_each_bucket(
        data_file_dict=data_file_dict,
        equality_delete_dict=equality_delete_dict,
        pos_delete_dict=pos_delete_dict,
    )
    iceberg_warehouse_bucket_name = params.iceberg_warehouse_bucket_name
    iceberg_namespace = params.iceberg_namespace
    iceberg_table_warehouse_prefix = construct_iceberg_table_prefix(
        iceberg_warehouse_bucket_name=iceberg_warehouse_bucket_name,
        table_name=table_name,
        iceberg_namespace=iceberg_namespace,
    )
    merge_keys = params.merge_keys
    # Using table identifier fields as merge keys if merge keys not provided
    if not merge_keys:
        identifier_fields_set = iceberg_table.schema().identifier_field_names()
        identifier_fields = list(identifier_fields_set)
    else:
        identifier_fields = merge_keys
    if len(identifier_fields) > 1:
        raise NotImplementedError(
            f"Multiple identifier fields lookup not supported yet."
        )
    convert_options_provider = functools.partial(
        task_resource_options_provider,
        resource_amount_provider=convert_resource_options_provider,
    )

    # TODO (zyiqin): max_parallel_data_file_download should be determined by memory requirement for each bucket.
    #  Specifically, for case when files for one bucket memory requirement exceed one worker node's memory limit, WITHOUT rebasing with larger hash bucket count,
    #  1. We can control parallel files to download by adjusting max_parallel_data_file_download.
    #  2. Implement two-layer converter tasks, with convert tasks to spin up child convert tasks.
    #  Note that approach 2 will ideally require shared object store to avoid download equality delete files * number of child tasks times.
    max_parallel_data_file_download = DEFAULT_MAX_PARALLEL_DATA_FILE_DOWNLOAD

    compact_small_files = params.compact_small_files
    position_delete_for_multiple_data_files = (
        params.position_delete_for_multiple_data_files
    )
    task_max_parallelism = params.task_max_parallelism

    def convert_input_provider(index, item):
        return {
            "convert_input": ConvertInput.of(
                files_for_each_bucket=item,
                convert_task_index=index,
                iceberg_table_warehouse_prefix=iceberg_table_warehouse_prefix,
                identifier_fields=identifier_fields,
                compact_small_files=compact_small_files,
                enforce_primary_key_uniqueness=enforce_primary_key_uniqueness,
                position_delete_for_multiple_data_files=position_delete_for_multiple_data_files,
                max_parallel_data_file_download=max_parallel_data_file_download,
            )
        }

    # Ray remote task: convert
    # Assuming that memory consume by each bucket doesn't exceed one node's memory limit.
    # TODO: Add split mechanism to split large buckets
    convert_tasks_pending = invoke_parallel(
        items=convert_input_files_for_all_buckets.items(),
        ray_task=convert,
        max_parallelism=task_max_parallelism,
        options_provider=convert_options_provider,
        kwargs_provider=convert_input_provider,
    )
    to_be_deleted_files_list = []
    to_be_added_files_dict_list = []
    convert_results = ray.get(convert_tasks_pending)
    for convert_result in convert_results:
        to_be_deleted_files_list.extend(convert_result[0].values())
        to_be_added_files_dict_list.append(convert_result[1])

    new_position_delete_files = parquet_files_dict_to_iceberg_data_files(
        io=iceberg_table.io,
        table_metadata=iceberg_table.metadata,
        files_dict_list=to_be_added_files_dict_list,
    )

    if not to_be_deleted_files_list:
        commit_append_snapshot(
            iceberg_table=iceberg_table,
            new_position_delete_files=new_position_delete_files,
        )
    else:
        commit_replace_snapshot(
            iceberg_table=iceberg_table,
            # equality_delete_files + data file that all rows are deleted
            to_be_deleted_files_list=to_be_deleted_files_list,
            new_position_delete_files=new_position_delete_files,
        )



from collections import defaultdict
import logging
from deltacat import logs
import pyarrow.parquet as pq

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def parquet_files_dict_to_iceberg_data_files(io, table_metadata, files_dict_list):
    from pyiceberg.io.pyarrow import (
        _check_pyarrow_schema_compatible,
        data_file_statistics_from_parquet_metadata,
        compute_statistics_plan,
        parquet_path_to_id_mapping,
    )
    from pyiceberg.manifest import (
        DataFile,
        DataFileContent,
        FileFormat,
    )

    data_file_content_type = DataFileContent.POSITION_DELETES
    iceberg_files = []
    schema = table_metadata.schema()
    for files_dict in files_dict_list:
        for partition_value, file_paths in files_dict.items():
            for file_path in file_paths:
                input_file = io.new_input(file_path)
                with input_file.open() as input_stream:
                    parquet_metadata = pq.read_metadata(input_stream)
                _check_pyarrow_schema_compatible(
                    schema, parquet_metadata.schema.to_arrow_schema()
                )

                statistics = data_file_statistics_from_parquet_metadata(
                    parquet_metadata=parquet_metadata,
                    stats_columns=compute_statistics_plan(
                        schema, table_metadata.properties
                    ),
                    parquet_column_mapping=parquet_path_to_id_mapping(schema),
                )

                data_file = DataFile(
                    content=data_file_content_type,
                    file_path=file_path,
                    file_format=FileFormat.PARQUET,
                    partition=partition_value,
                    # partition=Record(**{"pk": "111", "bucket": 2}),
                    file_size_in_bytes=len(input_file),
                    sort_order_id=None,
                    spec_id=table_metadata.default_spec_id,
                    equality_ids=None,
                    key_metadata=None,
                    **statistics.to_serialized_dict(),
                )
                iceberg_files.append(data_file)
    return iceberg_files


def fetch_all_bucket_files(table):
    # step 1: filter manifests using partition summaries
    # the filter depends on the partition spec used to write the manifest file, so create a cache of filters for each spec id
    from pyiceberg.typedef import (
        KeyDefaultDict,
    )

    data_scan = table.scan()
    snapshot = data_scan.snapshot()
    if not snapshot:
        return iter([])
    manifest_evaluators = KeyDefaultDict(data_scan._build_manifest_evaluator)

    manifests = [
        manifest_file
        for manifest_file in snapshot.manifests(data_scan.io)
        if manifest_evaluators[manifest_file.partition_spec_id](manifest_file)
    ]

    # step 2: filter the data files in each manifest
    # this filter depends on the partition spec used to write the manifest file
    from pyiceberg.expressions.visitors import _InclusiveMetricsEvaluator
    from pyiceberg.types import (
        strtobool,
    )
    from pyiceberg.table import _min_sequence_number, _open_manifest
    from pyiceberg.utils.concurrent import ExecutorFactory
    from itertools import chain
    from pyiceberg.manifest import DataFileContent

    partition_evaluators = KeyDefaultDict(data_scan._build_partition_evaluator)
    metrics_evaluator = _InclusiveMetricsEvaluator(
        data_scan.table_metadata.schema(),
        data_scan.row_filter,
        data_scan.case_sensitive,
        strtobool(data_scan.options.get("include_empty_files", "false")),
    ).eval

    min_sequence_number = _min_sequence_number(manifests)

    # {"bucket_index": List[DataFile]}
    data_entries = defaultdict(list)
    equality_data_entries = defaultdict(list)
    positional_delete_entries = defaultdict(list)

    executor = ExecutorFactory.get_or_create()
    for manifest_entry in chain(
        *executor.map(
            lambda args: _open_manifest(*args),
            [
                (
                    data_scan.io,
                    manifest,
                    partition_evaluators[manifest.partition_spec_id],
                    metrics_evaluator,
                )
                for manifest in manifests
                if data_scan._check_sequence_number(min_sequence_number, manifest)
            ],
        )
    ):
        data_file = manifest_entry.data_file
        file_sequence_number = manifest_entry.sequence_number
        data_file_tuple = (file_sequence_number, data_file)
        partition_value = data_file.partition
        if data_file.content == DataFileContent.DATA:
            data_entries[partition_value].append(data_file_tuple)
        if data_file.content == DataFileContent.POSITION_DELETES:
            positional_delete_entries[partition_value].append(data_file_tuple)
        elif data_file.content == DataFileContent.EQUALITY_DELETES:
            equality_data_entries[partition_value].append(data_file_tuple)
        else:
            logger.warning(
                f"Unknown DataFileContent ({data_file.content}): {manifest_entry}"
            )
    return data_entries, equality_data_entries, positional_delete_entries



import logging
from deltacat.compute.merge_on_read.model.merge_on_read_params import MergeOnReadParams
from deltacat.storage.model.types import DistributedDataset
from deltacat.types.media import TableType, DistributedDatasetType
from deltacat.compute.merge_on_read.utils.delta import create_df_from_all_deltas
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def merge(params: MergeOnReadParams, **kwargs) -> DistributedDataset:
    """
    Merges the given deltas and returns the result as distributed dataframe.
    It reads the deltas into the Daft dataframe and leverages operations supported
    by Daft to perform an efficient merge using Ray cluster.

    TODO(raghumdani): Perform actual merge.
    """

    delta_dfs = create_df_from_all_deltas(
        deltas=params.deltas,
        table_type=TableType.PYARROW,
        distributed_dataset_type=DistributedDatasetType.DAFT,
        reader_kwargs=params.reader_kwargs,
        deltacat_storage=params.deltacat_storage,
        deltacat_storage_kwargs=params.deltacat_storage_kwargs,
        **kwargs,
    )

    logger.info(f"Merging {len(delta_dfs)} delta dfs...")

    # TODO: This code should be optimized from daft side
    result = None
    for df in delta_dfs:
        if result is None:
            result = df
        else:
            result = result.concat(df)

    return result



from typing import List, Dict, Any, Optional, Union
from deltacat.storage.model.delta import Delta, DeltaLocator
from deltacat.storage.model.types import DistributedDataset
from deltacat.storage import (
    interface as unimplemented_deltacat_storage,
)
from deltacat.types.media import TableType, StorageType, DistributedDatasetType


def create_df_from_all_deltas(
    deltas: List[Union[Delta, DeltaLocator]],
    table_type: TableType,
    distributed_dataset_type: DistributedDatasetType,
    reader_kwargs: Optional[Dict[Any, Any]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[Any, Any]] = None,
    *args,
    **kwargs
) -> List[DistributedDataset]:  # type: ignore
    """
    This method creates a distributed dataset for each delta and returns their references.
    """

    if reader_kwargs is None:
        reader_kwargs = {}
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}

    df_list = []

    for delta in deltas:
        df = deltacat_storage.download_delta(
            delta_like=delta,
            table_type=table_type,
            distributed_dataset_type=distributed_dataset_type,
            storage_type=StorageType.DISTRIBUTED,
            **reader_kwargs,
            **deltacat_storage_kwargs
        )
        df_list.append(df)

    return df_list



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from typing import Any, Dict, List, Optional, Tuple
import pyarrow as pa

from deltacat.tests.compute.test_util_common import (
    PartitionKey,
)

from deltacat.storage import (
    Delta,
    DeltaType,
    Partition,
    PartitionLocator,
    Stream,
)
from deltacat.tests.compute.test_util_common import (
    create_src_table,
    create_destination_table,
    create_rebase_table,
)
import logging
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _add_deltas_to_partition(
    deltas_ingredients: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    partition: Optional[Partition],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> List[Optional[Delta], int]:
    import deltacat.tests.local_deltacat_storage as ds

    all_deltas_length = 0
    for (delta_data, delta_type, delete_parameters) in deltas_ingredients:
        staged_delta: Delta = ds.stage_delta(
            delta_data,
            partition,
            delta_type,
            entry_params=delete_parameters,
            **ds_mock_kwargs,
        )
        incremental_delta = ds.commit_delta(
            staged_delta,
            **ds_mock_kwargs,
        )
        all_deltas_length += len(delta_data) if delta_data else 0
    return incremental_delta, all_deltas_length


def add_late_deltas_to_partition(
    late_deltas: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    source_partition: Optional[Partition],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> List[Optional[Delta], int]:
    return _add_deltas_to_partition(late_deltas, source_partition, ds_mock_kwargs)


def create_incremental_deltas_on_source_table(
    source_namespace: str,
    source_table_name: str,
    source_table_version: str,
    source_table_stream: Stream,
    partition_values_param,
    incremental_deltas: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    ds_mock_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[PartitionLocator, Delta, int, bool]:
    import deltacat.tests.local_deltacat_storage as ds

    incremental_delta_length = 0
    is_delete = False
    src_partition: Partition = ds.get_partition(
        source_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    for (
        incremental_data,
        incremental_delta_type,
        incremental_delete_parameters,
    ) in incremental_deltas:
        if incremental_delta_type is DeltaType.DELETE:
            is_delete = True
        incremental_delta: Delta = ds.commit_delta(
            ds.stage_delta(
                incremental_data,
                src_partition,
                incremental_delta_type,
                entry_params=incremental_delete_parameters,
                **ds_mock_kwargs,
            ),
            **ds_mock_kwargs,
        )
        incremental_delta_length += len(incremental_data) if incremental_data else 0
    src_table_stream_after_committed_delta: Stream = ds.get_stream(
        source_namespace,
        source_table_name,
        source_table_version,
        **ds_mock_kwargs,
    )
    src_partition_after_committed_delta: Partition = ds.get_partition(
        src_table_stream_after_committed_delta.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    return (
        src_partition_after_committed_delta.locator,
        incremental_delta,
        incremental_delta_length,
        is_delete,
    )


def create_src_w_deltas_destination_plus_destination(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: pa.Table,
    input_delta_type: DeltaType,
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
    simulate_is_inplace: bool = False,
) -> Tuple[Stream, Stream, Optional[Stream], str, str, str]:
    import deltacat.tests.local_deltacat_storage as ds

    source_namespace, source_table_name, source_table_version = create_src_table(
        sort_keys, partition_keys, ds_mock_kwargs
    )

    source_table_stream: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        source_table_stream, partition_values, **ds_mock_kwargs
    )
    ds.commit_delta(
        ds.stage_delta(
            input_deltas, staged_partition, input_delta_type, **ds_mock_kwargs
        ),
        **ds_mock_kwargs,
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)
    source_table_stream_after_committed: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    destination_table_namespace: Optional[str] = None
    destination_table_name: Optional[str] = None
    destination_table_version: Optional[str] = None
    if not simulate_is_inplace:
        (
            destination_table_namespace,
            destination_table_name,
            destination_table_version,
        ) = create_destination_table(sort_keys, partition_keys, ds_mock_kwargs)
    else:
        # not creating a table as in-place
        destination_table_namespace = source_namespace
        destination_table_name = source_table_name
        destination_table_version = source_table_version

    destination_table_stream: Stream = ds.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )
    return (
        source_table_stream_after_committed,
        destination_table_stream,
        None,
        source_namespace,
        source_table_name,
        source_table_version,
    )


def create_src_w_deltas_destination_rebase_w_deltas_strategy(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: pa.Table,
    input_delta_type: DeltaType,
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Stream, Stream, Optional[Stream]]:
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.storage import Delta
    from deltacat.utils.common import current_time_ms

    last_stream_position = current_time_ms()
    source_namespace, source_table_name, source_table_version = create_src_table(
        sort_keys, partition_keys, ds_mock_kwargs
    )

    source_table_stream: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        source_table_stream, partition_values, **ds_mock_kwargs
    )
    staged_delta: Delta = ds.stage_delta(
        input_deltas, staged_partition, input_delta_type, **ds_mock_kwargs
    )
    staged_delta.locator.stream_position = last_stream_position
    ds.commit_delta(
        staged_delta,
        **ds_mock_kwargs,
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)
    source_table_stream_after_committed: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    # create the destination table
    (
        destination_table_namespace,
        destination_table_name,
        destination_table_version,
    ) = create_destination_table(sort_keys, partition_keys, ds_mock_kwargs)
    # create the rebase table
    (
        rebase_table_namespace,
        rebase_table_name,
        rebase_table_version,
    ) = create_rebase_table(sort_keys, partition_keys, ds_mock_kwargs)
    rebasing_table_stream: Stream = ds.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        rebasing_table_stream, partition_values, **ds_mock_kwargs
    )
    staged_delta: Delta = ds.stage_delta(
        input_deltas, staged_partition, **ds_mock_kwargs
    )
    staged_delta.locator.stream_position = last_stream_position
    ds.commit_delta(
        staged_delta,
        **ds_mock_kwargs,
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)

    # get streams
    # TODO: Add deltas to destination stream
    destination_table_stream: Stream = ds.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )
    rebased_stream_after_committed: Stream = ds.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )
    return (
        source_table_stream_after_committed,
        destination_table_stream,
        rebased_stream_after_committed,
    )


def multiple_rounds_create_src_w_deltas_destination_rebase_w_deltas_strategy(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: List[pa.Table],
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Stream, Stream, Optional[Stream], bool]:
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.storage import Partition, Stream

    source_namespace, source_table_name, source_table_version = create_src_table(
        sort_keys, partition_keys, ds_mock_kwargs
    )

    source_table_stream: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        source_table_stream, partition_values, **ds_mock_kwargs
    )
    is_delete = False
    input_delta_length = 0
    for (
        input_delta,
        input_delta_type,
        input_delta_parameters,
    ) in input_deltas:
        if input_delta_type is DeltaType.DELETE:
            is_delete = True
        staged_delta = ds.stage_delta(
            input_delta,
            staged_partition,
            input_delta_type,
            entry_params=input_delta_parameters,
            **ds_mock_kwargs,
        )
        ds.commit_delta(
            staged_delta,
            **ds_mock_kwargs,
        )
        input_delta_length += len(input_delta) if input_delta else 0
    ds.commit_partition(staged_partition, **ds_mock_kwargs)
    source_table_stream_after_committed: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    # create the destination table
    (
        destination_table_namespace,
        destination_table_name,
        destination_table_version,
    ) = create_destination_table(sort_keys, partition_keys, ds_mock_kwargs)
    # create the rebase table
    (
        rebase_table_namespace,
        rebase_table_name,
        rebase_table_version,
    ) = create_rebase_table(sort_keys, partition_keys, ds_mock_kwargs)
    rebasing_table_stream: Stream = ds.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        rebasing_table_stream, partition_values, **ds_mock_kwargs
    )
    input_delta_length = 0
    for (
        input_delta,
        input_delta_type,
        input_delta_parameters,
    ) in input_deltas:
        if input_delta_type is DeltaType.DELETE:
            is_delete = True
        staged_delta = ds.stage_delta(
            input_delta,
            staged_partition,
            input_delta_type,
            entry_params=input_delta_parameters,
            **ds_mock_kwargs,
        )
        ds.commit_delta(
            staged_delta,
            **ds_mock_kwargs,
        )
        input_delta_length += len(input_delta) if input_delta else 0
    ds.commit_partition(staged_partition, **ds_mock_kwargs)

    # get streams
    destination_table_stream: Stream = ds.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )
    rebased_stream_after_committed: Stream = ds.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )
    return (
        source_table_stream_after_committed,
        destination_table_stream,
        rebased_stream_after_committed,
        is_delete,
    )



from typing import Optional

from pyiceberg.catalog import Catalog
from deltacat.storage.model.scan.push_down import Pushdown
from deltacat.storage.model.scan.scan_plan import ScanPlan
from deltacat.storage.model.scan.scan_task import FileScanTask, DataFile
from deltacat.storage.util.scan_planner import ScanPlanner
from deltacat.storage.iceberg.impl import _try_load_iceberg_table


class IcebergScanPlanner(ScanPlanner):
    def __init__(self, catalog: Catalog):
        self.catalog = catalog

    def create_scan_plan(
        self,
        table_name: str,
        namespace: Optional[str] = None,
        pushdown: Optional[Pushdown] = None,
    ) -> ScanPlan:
        iceberg_table = _try_load_iceberg_table(
            self.catalog, namespace=namespace, table_name=table_name
        )
        file_scan_tasks = []
        # TODO: implement predicate pushdown to Iceberg
        for scan_task in iceberg_table.scan().plan_files():
            file_scan_tasks.append(FileScanTask([DataFile(scan_task.file.file_path)]))
        return ScanPlan(file_scan_tasks)



from typing import Optional, List
import uuid
from pyiceberg.table.snapshots import (
    Operation,
)
from pyiceberg.manifest import (
    DataFileContent,
    ManifestContent,
    ManifestEntry,
    ManifestEntryStatus,
    ManifestFile,
    write_manifest,
)
import itertools
from pyiceberg.utils.concurrent import ExecutorFactory
from pyiceberg.table.update.snapshot import UpdateSnapshot, _SnapshotProducer


class _ReplaceFiles(_SnapshotProducer["_ReplaceFiles"]):
    """Overwrites data from the table. This will produce an OVERWRITE snapshot.

    Data and delete files were added and removed in a logical overwrite operation.
    """

    def _existing_manifests(self) -> List[ManifestFile]:
        """Determine if there are any existing manifest files."""
        existing_files = []
        snapshot = self._transaction.table_metadata.current_snapshot()
        if snapshot:
            for manifest_file in snapshot.manifests(io=self._io):
                entries = manifest_file.fetch_manifest_entry(
                    io=self._io, discard_deleted=True
                )

                found_deleted_data_files = [
                    entry.data_file
                    for entry in entries
                    if entry.data_file in self._deleted_data_files
                ]

                if len(found_deleted_data_files) == 0:
                    existing_files.append(manifest_file)
                else:
                    # We have to replace the manifest file without the deleted data files
                    if any(
                        entry.data_file not in found_deleted_data_files
                        for entry in entries
                    ):
                        with write_manifest(
                            format_version=self._transaction.table_metadata.format_version,
                            spec=self._transaction.table_metadata.specs()[
                                manifest_file.partition_spec_id
                            ],
                            schema=self._transaction.table_metadata.schema(),
                            output_file=self.new_manifest_output(),
                            snapshot_id=self._snapshot_id,
                        ) as writer:
                            [
                                writer.add_entry(
                                    ManifestEntry(
                                        status=ManifestEntryStatus.EXISTING,
                                        snapshot_id=entry.snapshot_id,
                                        sequence_number=entry.sequence_number,
                                        file_sequence_number=entry.file_sequence_number,
                                        data_file=entry.data_file,
                                    )
                                )
                                for entry in entries
                                if entry.data_file not in found_deleted_data_files
                            ]
                        existing_files.append(writer.to_manifest_file())
        return existing_files

    def _deleted_entries(self) -> List[ManifestEntry]:
        """To determine if we need to record any deleted entries.

        With a full overwrite all the entries are considered deleted.
        With partial overwrites we have to use the predicate to evaluate
        which entries are affected.
        """
        if self._parent_snapshot_id is not None:
            previous_snapshot = self._transaction.table_metadata.snapshot_by_id(
                self._parent_snapshot_id
            )
            if previous_snapshot is None:
                # This should never happen since you cannot overwrite an empty table
                raise ValueError(
                    f"Could not find the previous snapshot: {self._parent_snapshot_id}"
                )

            executor = ExecutorFactory.get_or_create()

            def _get_entries(manifest: ManifestFile) -> List[ManifestEntry]:
                return [
                    ManifestEntry(
                        status=ManifestEntryStatus.DELETED,
                        snapshot_id=entry.snapshot_id,
                        sequence_number=entry.sequence_number,
                        file_sequence_number=entry.file_sequence_number,
                        data_file=entry.data_file,
                    )
                    for entry in manifest.fetch_manifest_entry(
                        self._io, discard_deleted=True
                    )
                    if entry.data_file.content == DataFileContent.DATA
                    and entry.data_file in self._deleted_data_files
                ]

            list_of_entries = executor.map(
                _get_entries, previous_snapshot.manifests(self._io)
            )
            return list(itertools.chain(*list_of_entries))
        else:
            return []


def replace(
    self,
    commit_uuid: Optional[uuid.UUID] = None,
    using_starting_sequence: Optional[bool] = False,
) -> _ReplaceFiles:
    return _ReplaceFiles(
        commit_uuid=commit_uuid,
        operation=Operation.REPLACE
        if self._transaction.table_metadata.current_snapshot() is not None
        else Operation.APPEND,
        transaction=self._transaction,
        io=self._io,
        snapshot_properties=self._snapshot_properties,
        using_starting_sequence=using_starting_sequence,
    )


UpdateSnapshot.replace = replace


def commit_replace_snapshot(
    iceberg_table, to_be_deleted_files_list, new_position_delete_files
):
    tx = iceberg_table.transaction()
    snapshot_properties = {}
    commit_uuid = uuid.uuid4()
    update_snapshot = tx.update_snapshot(snapshot_properties=snapshot_properties)
    replace_snapshot = replace(
        self=update_snapshot, commit_uuid=commit_uuid, using_starting_sequence=False
    )
    for to_be_deleted_file in to_be_deleted_files_list:
        replace_snapshot.append_data_file(to_be_deleted_file)
    for to_be_added_file in new_position_delete_files:
        replace_snapshot.delete_data_file(to_be_added_file)
    replace_snapshot._commit()
    tx.commit_transaction()


def append_delete_files_override(update_snapshot):
    commit_uuid = uuid.uuid4()
    return _AppendDeleteFilesOverride(
        commit_uuid=commit_uuid,
        operation=Operation.APPEND,
        transaction=update_snapshot._transaction,
        io=update_snapshot._io,
        snapshot_properties=update_snapshot._snapshot_properties,
    )


class _AppendDeleteFilesOverride(_SnapshotProducer):
    def _manifests(self):
        def _write_added_manifest():
            if self._added_data_files:
                with write_manifest(
                    format_version=self._transaction.table_metadata.format_version,
                    spec=self._transaction.table_metadata.spec(),
                    schema=self._transaction.table_metadata.schema(),
                    output_file=self.new_manifest_output(),
                    snapshot_id=self._snapshot_id,
                ) as writer:
                    for data_file in self._added_data_files:
                        writer.add(
                            ManifestEntry(
                                status=ManifestEntryStatus.ADDED,
                                snapshot_id=self._snapshot_id,
                                sequence_number=None,
                                file_sequence_number=None,
                                data_file=data_file,
                            )
                        )
                        writer.content = self.writer_content
                return [writer.to_manifest_file()]
            else:
                return []

        executor = ExecutorFactory.get_or_create()

        added_manifests = executor.submit(_write_added_manifest)
        existing_manifests = executor.submit(self._existing_manifests)

        return self._process_manifests(
            added_manifests.result() + existing_manifests.result()
        )

    def writer_content(self):
        return ManifestContent.DELETES

    def _existing_manifests(self) -> List[ManifestFile]:
        """To determine if there are any existing manifest files.

        A fast append will add another ManifestFile to the ManifestList.
        All the existing manifest files are considered existing.
        """
        existing_manifests = []

        if self._parent_snapshot_id is not None:
            previous_snapshot = self._transaction.table_metadata.snapshot_by_id(
                self._parent_snapshot_id
            )

            if previous_snapshot is None:
                raise ValueError(
                    f"Snapshot could not be found: {self._parent_snapshot_id}"
                )

            for manifest in previous_snapshot.manifests(io=self._io):
                if (
                    manifest.has_added_files()
                    or manifest.has_existing_files()
                    or manifest.added_snapshot_id == self._snapshot_id
                ):
                    existing_manifests.append(manifest)

        return existing_manifests

    def _deleted_entries(self) -> List[ManifestEntry]:
        """To determine if we need to record any deleted manifest entries.

        In case of an append, nothing is deleted.
        """
        return []


def commit_append_snapshot(iceberg_table, new_position_delete_files):
    with iceberg_table.transaction() as tx:
        if iceberg_table.metadata.name_mapping() is None:
            tx.set_properties(
                **{
                    "schema.name-mapping.default": tx.table_metadata.schema().name_mapping.model_dump_json()
                }
            )
        with append_delete_files_override(tx.update_snapshot()) as append_snapshot:
            if new_position_delete_files:
                for data_file in new_position_delete_files:
                    append_snapshot.append_data_file(data_file)



import pyarrow as pa
import pyarrow.compute as pc
import deltacat.compute.converter.utils.iceberg_columns as sc
from deltacat.compute.converter.utils.io import (
    download_data_table_and_append_iceberg_columns,
)


def dedupe_data_files(
    data_file_to_dedupe,
    identify_column_name_concatenated,
    identifier_columns,
    merge_sort_column,
):
    data_file_table = []

    # Sort data files by file sequence number first
    data_file_to_dedupe = sorted(data_file_to_dedupe, key=lambda f: f[0])
    for file_tuple in data_file_to_dedupe:
        sequence_number = file_tuple[0]
        data_file = file_tuple[1]
        data_file_to_dedupe_table = download_data_table_and_append_iceberg_columns(
            file=data_file,
            columns_to_download=identifier_columns,
            additional_columns_to_append=[
                sc._FILE_PATH_COLUMN_NAME,
                sc._ORDERED_RECORD_IDX_COLUMN_NAME,
            ],
            sequence_number=sequence_number,
        )
        data_file_table.append(data_file_to_dedupe_table)

    final_data_to_dedupe = pa.concat_tables(data_file_table)

    record_idx_iterator = iter(range(len(final_data_to_dedupe)))

    # Append global record index to used as aggregate column
    final_data_to_dedupe = sc.append_global_record_idx_column(
        final_data_to_dedupe, record_idx_iterator
    )

    final_data_table_indices = final_data_to_dedupe.group_by(
        identify_column_name_concatenated, use_threads=False
    ).aggregate([(sc._GLOBAL_RECORD_IDX_COLUMN_NAME, "max")])

    pos_delete_indices = pc.invert(
        pc.is_in(
            final_data_to_dedupe[sc._GLOBAL_RECORD_IDX_COLUMN_NAME],
            value_set=final_data_table_indices[
                f"{sc._GLOBAL_RECORD_IDX_COLUMN_NAME}_max"
            ],
        )
    )

    final_data_table_to_delete = final_data_to_dedupe.filter(pos_delete_indices)

    final_data_table_to_delete = final_data_table_to_delete.drop(
        [identify_column_name_concatenated, sc._GLOBAL_RECORD_IDX_COLUMN_NAME]
    )
    return final_data_table_to_delete



import unittest
import pytest
import ray
import tempfile
import shutil
import uuid
from unittest import mock
import os

from deltacat.catalog import CatalogProperties
from pyiceberg.catalog import Catalog as IcebergCatalog

from deltacat.catalog.model.catalog import (
    Catalog,
    init,
    get_catalog,
    put_catalog,
    is_initialized,
)
from deltacat.catalog.iceberg.iceberg_catalog_config import IcebergCatalogConfig

from pyiceberg.catalog import CatalogType


# Test module to mock a catalog implementation
class MockCatalogImpl:
    @staticmethod
    def initialize(*args, **kwargs):
        # Return some state that the catalog would normally maintain
        return {"initialized": True, "args": args, "kwargs": kwargs}


@pytest.fixture(scope="function")
def reset_catalogs_ray_actor():
    """
    Setup and teardown for Ray environment for tests.

    This will kill the actor all_catalogs, essentially wiping global state for catalogs

    NOTE: tests using this fixture must be run serially. As of April 7 2025, the unit test suite had various
      failures if run in parallel, in part because the state of all_catalogs in ray is shared across tests.

    NOTE: when using this fixture, ensure you pass ray_init_args={"ignore_reinit_error": True} into all
        functions which may re-initialize ray. This is because the production code checks the all_catalogs actor
        in order to determine whether it needs to initialize Ray
    """
    # Reset the global catalog_actor state before each test
    import deltacat.catalog.model.catalog as catalog_module

    # Initialize Ray if not already initialized
    if not ray.is_initialized():
        ray.init(ignore_reinit_error=True)
    yield

    # Clean up the actor if it exists
    if catalog_module.all_catalogs is not None:
        try:
            ray.kill(catalog_module.all_catalogs)
        except Exception:
            pass
        finally:
            catalog_module.all_catalogs = None


class TestCatalog(unittest.TestCase):
    """Tests for the Catalog class itself, without Ray initialization."""

    def test_catalog_constructor(self):
        """Test that the Catalog constructor correctly initializes with the given implementation."""
        catalog = Catalog(impl=MockCatalogImpl)

        self.assertEqual(catalog.impl, MockCatalogImpl)

        # Check that inner state was correctly initialized
        # This just asserts that kwargs were plumbed through from Catalog constructor
        self.assertTrue(catalog.inner["initialized"])
        self.assertEqual(catalog.inner["args"], ())
        self.assertEqual(catalog.inner["kwargs"], {})

    def test_iceberg_factory_method(self):
        """Test the iceberg factory method correctly creates an Iceberg catalog."""
        # Create a mock for the Iceberg catalog module
        with mock.patch(
            "deltacat.catalog.model.catalog.IcebergCatalog"
        ) as mock_iceberg_catalog:
            # Configure the mock to return a known value when initialize is called
            mock_iceberg_catalog.initialize.return_value = {"iceberg": True}

            # Create an Iceberg catalog config and invoke iceberg factory method
            config = IcebergCatalogConfig(type=CatalogType.IN_MEMORY, properties={})
            catalog = Catalog.iceberg(config)

            # Check that the implementation is set to iceberg_catalog
            self.assertEqual(catalog.impl, mock_iceberg_catalog)
            # Check that the inner state is set to the output of initialize
            self.assertEqual(catalog.inner, {"iceberg": True})


class TestCatalogsIntegration:
    """Integration tests for Default catalog functionality."""

    temp_dir = None

    @classmethod
    def setup_class(cls):
        cls.temp_dir = tempfile.mkdtemp()
        # Other tests are going to have initialized ray catalog. Initialize here to ensure
        # that when this test class is run individuall it mimicks running with other tests
        catalog = Catalog(impl=MockCatalogImpl)
        init(
            catalog,
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

    @classmethod
    def teardown_class(cls):
        if cls.temp_dir and os.path.exists(cls.temp_dir):
            shutil.rmtree(cls.temp_dir)

    def test_init_single_catalog(self, reset_catalogs_ray_actor):
        """Test initializing a single catalog."""

        catalog = Catalog(impl=MockCatalogImpl)

        # Initialize with a single catalog and Ray init args including the namespace
        init(
            catalog,
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        assert is_initialized()

        # Get the default catalog and check it's the same one we initialized with
        retrieved_catalog = get_catalog()
        assert retrieved_catalog.impl == MockCatalogImpl
        assert retrieved_catalog.inner["initialized"]

    def test_init_multiple_catalogs(self, reset_catalogs_ray_actor):
        """Test initializing multiple catalogs."""
        # Create catalogs
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)

        # Initialize with multiple catalogs and Ray init args including the namespace
        catalogs_dict = {"catalog1": catalog1, "catalog2": catalog2}
        init(
            catalogs_dict,
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        assert is_initialized()

        # Get catalogs by name and check they're the same ones we initialized with
        retrieved_catalog1 = get_catalog("catalog1")
        assert retrieved_catalog1.impl == MockCatalogImpl
        assert retrieved_catalog1.inner["kwargs"]["id"] == 1

        retrieved_catalog2 = get_catalog("catalog2")
        assert retrieved_catalog2.impl == MockCatalogImpl
        assert retrieved_catalog2.inner["kwargs"]["id"] == 2

    def test_init_with_default_catalog_name(self, reset_catalogs_ray_actor):
        """Test initializing with a specified default catalog name."""
        # Create catalogs
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)

        # Initialize with multiple catalogs and specify a default
        catalogs_dict = {"catalog1": catalog1, "catalog2": catalog2}
        init(
            catalogs_dict,
            default="catalog2",
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        # Get the default catalog and check it's catalog2
        default_catalog = get_catalog()
        assert default_catalog.impl == MockCatalogImpl
        assert default_catalog.inner["kwargs"]["id"] == 2

    def test_put_catalog(self, reset_catalogs_ray_actor):
        """Test adding a catalog after initialization."""
        # Initialize with a single catalog
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)
        init(
            {"catalog1": catalog1},
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        # Add a second catalog
        put_catalog("catalog2", catalog2)

        # Check both catalogs are available
        retrieved_catalog1 = get_catalog("catalog1")
        assert retrieved_catalog1.inner["kwargs"]["id"] == 1

        retrieved_catalog2 = get_catalog("catalog2")
        assert retrieved_catalog2.inner["kwargs"]["id"] == 2

    def test_put_catalog_that_already_exists(self, reset_catalogs_ray_actor):
        catalog = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)
        put_catalog(
            "test_catalog",
            catalog,
            id=1,
            ray_init_args={"ignore_reinit_error": True},
        )

        # Try to add another catalog with the same name. Should not error
        put_catalog(
            "test_catalog",
            catalog2,
            ray_init_args={"ignore_reinit_error": True},
        )

        retrieved_catalog = get_catalog("test_catalog")
        assert retrieved_catalog.inner["kwargs"]["id"] == 2

        # If fail_if_exists, put call should fail
        with pytest.raises(ValueError):
            put_catalog(
                "test_catalog",
                catalog,
                ray_init_args={"ignore_reinit_error": True},
                fail_if_exists=True,
            )

    def test_get_catalog_nonexistent(self, reset_catalogs_ray_actor):
        """Test that trying to get a nonexistent catalog raises an error."""
        # Initialize with a catalog
        catalog = Catalog(impl=MockCatalogImpl)
        init(
            {"test_catalog": catalog},
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        # Try to get a nonexistent catalog
        with pytest.raises(ValueError):
            get_catalog("nonexistent")

    def test_get_catalog_no_default(self, reset_catalogs_ray_actor):
        """Test that trying to get the default catalog when none is set raises an error."""
        # Initialize with multiple catalogs but no default
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)
        init(
            {"catalog1": catalog1, "catalog2": catalog2},
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        # Try to get the default catalog
        with pytest.raises(ValueError):
            get_catalog()

    def test_default_catalog_initialization(self, reset_catalogs_ray_actor):
        """Test that a Default catalog can be initialized and accessed using the factory method."""
        from deltacat.catalog.model.properties import CatalogProperties

        catalog_name = str(uuid.uuid4())

        # Create the catalog properties
        config = CatalogProperties(root=self.temp_dir)

        # Create the catalog using the factory method
        catalog = Catalog.default(config)

        # Initialize DeltaCAT with this catalog
        init(
            {catalog_name: catalog},
            ray_init_args={"ignore_reinit_error": True},
            **{"force_reinitialize": True},
        )

        # Retrieve the catalog and verify it's the same one
        retrieved_catalog = get_catalog(catalog_name)
        assert retrieved_catalog.impl.__name__ == "deltacat.catalog.main.impl"
        assert isinstance(retrieved_catalog.inner, CatalogProperties)
        assert retrieved_catalog.inner.root == self.temp_dir

    def test_default_catalog_initialization_from_kwargs(self, reset_catalogs_ray_actor):

        catalog_name = str(uuid.uuid4())
        # Initialize DeltaCAT with this catalog
        from deltacat.catalog.main import impl as DeltacatCatalog

        put_catalog(
            catalog_name,
            Catalog(DeltacatCatalog, **{"root": "test_root"}),
            ray_init_args={"ignore_reinit_error": True},
        )

        # Retrieve the catalog and verify it's the same one
        retrieved_catalog = get_catalog(catalog_name)
        assert retrieved_catalog.impl.__name__ == "deltacat.catalog.main.impl"
        assert isinstance(retrieved_catalog.inner, CatalogProperties)
        assert retrieved_catalog.inner.root == "test_root"

    def test_iceberg_catalog_initialization(self, reset_catalogs_ray_actor):
        """Test that an Iceberg catalog can be initialized and accessed."""
        catalog_name = str(uuid.uuid4())

        # Create the Iceberg catalog config
        config = IcebergCatalogConfig(
            type=CatalogType.IN_MEMORY, properties={"warehouse": self.temp_dir}
        )

        # Create the catalog using the factory method
        catalog = Catalog.iceberg(config)

        put_catalog(catalog_name, catalog, ray_init_args={"ignore_reinit_error": True})

        # Retrieve the catalog and verify it's the same one
        retrieved_catalog = get_catalog(catalog_name)
        assert retrieved_catalog.impl.__name__ == "deltacat.catalog.iceberg.impl"
        assert isinstance(retrieved_catalog.inner, IcebergCatalog)



import pytest
import os
from moto import mock_s3
import boto3
from boto3.resources.base import ServiceResource
from deltacat.compute.compactor.utils.round_completion_file import (
    read_round_completion_file,
    write_round_completion_file,
)
from deltacat.tests.compute.test_util_common import get_test_partition_locator
from deltacat.compute.compactor import RoundCompletionInfo

RCF_BUCKET_NAME = "rcf-bucket"


@pytest.fixture(autouse=True, scope="module")
def mock_aws_credential():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_ID"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    yield


@pytest.fixture(autouse=True, scope="module")
def s3_resource(mock_aws_credential):
    with mock_s3():
        yield boto3.resource("s3")


@pytest.fixture(autouse=True, scope="function")
def setup_compaction_artifacts_s3_bucket(s3_resource: ServiceResource):
    s3_resource.create_bucket(
        ACL="authenticated-read",
        Bucket=RCF_BUCKET_NAME,
    )
    yield
    s3_resource.Bucket(RCF_BUCKET_NAME).objects.all().delete()


class TestReadWriteRoundCompletionFile:
    def test_read_when_rcf_written_without_destination(self):
        """
        This test case tests the backward compatibility by successfully
        reading the previously written rcf.
        """

        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        rcf_url = write_round_completion_file(
            RCF_BUCKET_NAME, source_locator, None, expected_rcf
        )

        rcf = read_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator
        )

        assert (
            rcf_url == "s3://rcf-bucket/f9829af39770d904dbb811bd8f4e886dd307f507.json"
        )
        assert rcf == expected_rcf

    def test_read_when_rcf_written_with_destination(self):
        """
        This test case tests the backward compatibility by successfully
        reading the previously written rcf.
        """

        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        rcf_url = write_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator, expected_rcf
        )

        rcf = read_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator
        )

        assert (
            rcf_url
            == "s3://rcf-bucket/f9829af39770d904dbb811bd8f4e886dd307f507/e9939deadc091b3289a2eb0ca56b1ba86b9892f4.json"
        )
        assert rcf == expected_rcf

    def test_read_without_destination_when_rcf_written_with_destination(self):
        """
        This test case tests the backward compatibility by successfully
        reading the previously written rcf.
        """

        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        write_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator, expected_rcf
        )

        rcf = read_round_completion_file(RCF_BUCKET_NAME, source_locator, None)

        assert rcf is None

    def test_read_without_destination_when_rcf_written_without_destination(self):
        """
        This test case tests the backward compatibility by successfully
        reading the previously written rcf.
        """

        source_locator = get_test_partition_locator("source")

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        write_round_completion_file(RCF_BUCKET_NAME, source_locator, None, expected_rcf)

        rcf = read_round_completion_file(RCF_BUCKET_NAME, source_locator, None)

        assert rcf == expected_rcf

    def test_read_when_rcf_written_both_with_and_without_destination(self):
        """
        This test case tests the backward compatibility by successfully
        reading the previously written rcf.
        """

        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator("destination")

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        expected_rcf_2 = RoundCompletionInfo.of(
            high_watermark=1223,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=1233,
        )

        write_round_completion_file(RCF_BUCKET_NAME, source_locator, None, expected_rcf)

        write_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator, expected_rcf_2
        )

        rcf = read_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator
        )

        assert rcf == expected_rcf_2

    def test_read_when_none_destination_partition_id(self):

        source_locator = get_test_partition_locator("source")
        destination_locator = get_test_partition_locator(None)

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        write_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator, expected_rcf
        )

        rcf = read_round_completion_file(
            RCF_BUCKET_NAME, source_locator, destination_locator
        )

        assert rcf == expected_rcf

    def test_write_when_custom_url_is_passed(self):
        """
        This test case tests the backward compatibility by successfully
        reading the previously written rcf.
        """

        source_locator = get_test_partition_locator("source")

        expected_rcf = RoundCompletionInfo.of(
            high_watermark=122,
            compacted_delta_locator={},
            compacted_pyarrow_write_result={},
            sort_keys_bit_width=12,
        )

        completion_file_s3_url = f"s3://{RCF_BUCKET_NAME}/test.json"
        rcf_url = write_round_completion_file(
            RCF_BUCKET_NAME,
            source_locator,
            None,
            expected_rcf,
            completion_file_s3_url=completion_file_s3_url,
        )

        rcf = read_round_completion_file(RCF_BUCKET_NAME, source_locator, None)

        assert rcf_url == completion_file_s3_url
        assert rcf is None



import os
import logging
import argparse
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def store_cli_args_in_os_environ(script_args_list=[]):
    parser = argparse.ArgumentParser()
    for args, kwargs in script_args_list:
        parser.add_argument(*args, **kwargs)
    args = parser.parse_args()
    print(f"Command Line Arguments: {args}")
    os.environ.update(vars(args))



from __future__ import annotations
from typing import Optional, Any

import pyarrow
from deltacat.constants import DELTACAT_ROOT

from deltacat.utils.filesystem import resolve_path_and_filesystem


def get_catalog_properties(
    *args,
    catalog: Optional[CatalogProperties] = None,
    inner: Optional[CatalogProperties] = None,
    **kwargs,
) -> CatalogProperties:
    """
    Helper function to fetch CatalogProperties instance. You are meant to call this by providing your functions
    kwargs, OR to directly pass through CatalogProperty configuration keys like "root" in kwargs.

    This will look for a CatalogProperty value in the kwargs "catalog" or "inner". If these are found, it returns
    the CatalogProperty value under that kwarg. Otherwise, it will pass through kwargs to the CatalogProperties
    constructor.
    """
    properties = catalog if catalog is not None else inner
    if properties is not None and isinstance(properties, CatalogProperties):
        return properties
    elif properties is not None and not isinstance(properties, CatalogProperties):
        raise ValueError(
            f"Expected catalog properties of type {CatalogProperties.__name__} "
            f"but found {type(properties)}."
        )
    else:
        return CatalogProperties(**kwargs)


class CatalogProperties:
    """
    DeltaCAT catalog properties used to deterministically resolve a durable
    DeltaCAT catalog instance. Properties are set from system environment
    variables unless explicit overrides are provided during initialization.

    Catalog and storage APIs rely on the property catalog to retrieve durable state about the catalog they're
    working against.

    Attributes:
        root (str): URI string The root path where catalog metadata and data
            files are stored. Root is determined (in prededence order) by:
            1. check kwargs for "root"
            2. check env variable "DELTACAT_ROOT"
            3. default to ${cwd}/.deltacat

        filesystem: The filesystem implementation that should be used for
            reading/writing files. If None, a filesystem will be inferred from
            the catalog root path.

        storage: Storage class implementation (overrides default filesystem storage impl)
    """

    def __init__(
        self,
        root: Optional[str] = None,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
        storage=None,
        *args,
        **kwargs,
    ):
        """
        Initialize a CatalogProperties instance.

        Args:
            root: A single directory path that serves as the catalog root dir.
            filesystem: The filesystem implementation that should be used for
                reading these files. If None, a filesystem will be inferred.
                If not None, the provided filesystem will still be validated
                against the provided path to ensure compatibility.
        """
        # set root, using precedence rules described in pydoc
        if root is None:
            # Check environment variables
            # This is set or defaulted in constants.py
            root = DELTACAT_ROOT
            if root is None:
                raise ValueError(
                    "Expected environment variable DELTACAT_ROOT to be set or defaulted"
                )

        resolved_root, resolved_filesystem = resolve_path_and_filesystem(
            path=root,
            filesystem=filesystem,
        )
        self._root = resolved_root
        self._filesystem = resolved_filesystem
        self._storage = storage

    @property
    def root(self) -> str:
        return self._root

    @property
    def filesystem(self) -> Optional[pyarrow.fs.FileSystem]:
        return self._filesystem

    @property
    def storage(self) -> Optional[Any]:
        """
        Return overridden storage impl, if any
        """
        return self._storage

    def __str__(self):
        return (
            f"{self.__class__.__name__}(root={self.root}, filesystem={self.filesystem})"
        )

    def __repr__(self):
        return self.__str__()



import pytest
import ray
from typing import List
from pyiceberg.catalog.rest import RestCatalog
from pyiceberg.expressions import EqualTo
from pyiceberg.schema import Schema
from pyiceberg.types import (
    NestedField,
    StringType,
    LongType,
)
from pyiceberg.partitioning import PartitionSpec, PartitionField
from pyiceberg.transforms import IdentityTransform
import pyarrow as pa

from deltacat.compute.converter.steps.convert import convert
from deltacat.compute.converter.model.convert_input import ConvertInput
from deltacat.compute.converter.pyiceberg.overrides import (
    fetch_all_bucket_files,
    parquet_files_dict_to_iceberg_data_files,
)
from collections import defaultdict
from deltacat.compute.converter.utils.converter_session_utils import (
    group_all_files_to_each_bucket,
)
from deltacat.tests.compute.converter.utils import (
    get_s3_file_system,
    drop_table_if_exists,
)
from deltacat.compute.converter.pyiceberg.update_snapshot_overrides import (
    commit_append_snapshot,
)


def run_spark_commands(spark, sqls: List[str]) -> None:
    for sql in sqls:
        spark.sql(sql)


@pytest.mark.integration
def test_pyiceberg_spark_setup_sanity(spark, session_catalog: RestCatalog) -> None:
    """
    This Test was copied over from Pyiceberg integ test: https://github.com/apache/iceberg-python/blob/main/tests/integration/test_deletes.py#L62
    First sanity check to ensure all integration with Pyiceberg and Spark are working as expected.
    """
    identifier = "default.table_partitioned_delete"

    run_spark_commands(
        spark,
        [
            f"DROP TABLE IF EXISTS {identifier}",
            f"""
            CREATE TABLE {identifier} (
                number_partitioned  int,
                number              int
            )
            USING iceberg
            PARTITIONED BY (number_partitioned)
            TBLPROPERTIES('format-version' = 2)
        """,
            f"""
            INSERT INTO {identifier} VALUES (10, 20), (10, 30)
        """,
            f"""
            INSERT INTO {identifier} VALUES (11, 20), (11, 30)
        """,
        ],
    )

    tbl = session_catalog.load_table(identifier)
    tbl.delete(EqualTo("number_partitioned", 10))

    # No overwrite operation
    assert [snapshot.summary.operation.value for snapshot in tbl.snapshots()] == [
        "append",
        "append",
        "delete",
    ]
    assert tbl.scan().to_arrow().to_pydict() == {
        "number_partitioned": [11, 11],
        "number": [20, 30],
    }


@pytest.mark.integration
def test_spark_position_delete_production_sanity(
    spark, session_catalog: RestCatalog
) -> None:
    """
    Sanity test to ensure Spark position delete production is successful with `merge-on-read` spec V2.
    Table has two partition levels. 1. BucketTransform on primary key
    """
    identifier = "default.table_spark_position_delete_production_sanity"

    run_spark_commands(
        spark,
        [
            f"DROP TABLE IF EXISTS {identifier}",
            f"""
            CREATE TABLE {identifier} (
                number_partitioned INT,
                primary_key STRING
            )
            USING iceberg
            PARTITIONED BY (bucket(3, primary_key), number_partitioned)
            TBLPROPERTIES(
                'format-version' = 2,
                'write.delete.mode'='merge-on-read',
                'write.update.mode'='merge-on-read',
                'write.merge.mode'='merge-on-read'
            )
            """,
            f"""
            INSERT INTO {identifier} VALUES (0, 'pk1'), (0, 'pk2'), (0, 'pk3')
            """,
            f"""
            INSERT INTO {identifier} VALUES (1, 'pk1'), (1, 'pk2'), (1, 'pk3')
            """,
        ],
    )

    run_spark_commands(
        spark,
        [
            f"""
                DELETE FROM {identifier} WHERE primary_key in ("pk1")
            """,
        ],
    )

    tbl = session_catalog.load_table(identifier)
    tbl.refresh()

    assert [snapshot.summary.operation.value for snapshot in tbl.snapshots()] == [
        "append",
        "append",
        "delete",
    ]

    assert tbl.scan().to_arrow().to_pydict() == {
        "number_partitioned": [1, 1, 0, 0],
        "primary_key": ["pk2", "pk3", "pk2", "pk3"],
    }


@pytest.mark.integration
def test_converter_drop_duplicates_success(
    spark, session_catalog: RestCatalog, setup_ray_cluster, mocker
) -> None:
    """
    Test for convert compute remote function happy case. Download file results are mocked.
    """

    # 1. Create Iceberg table
    namespace = "default"
    table_name = "table_converter_ray_pos_delete_drop_duplicates_compute"
    identifier = f"{namespace}.{table_name}"

    schema = Schema(
        NestedField(
            field_id=1, name="number_partitioned", field_type=LongType(), required=False
        ),
        NestedField(
            field_id=2, name="primary_key", field_type=StringType(), required=False
        ),
        # Explicitly define "file_path" and "pos" for assertion of deterministic record after dedupe
        NestedField(
            field_id=2147483546,
            name="file_path",
            field_type=StringType(),
            required=False,
        ),
        NestedField(
            field_id=2147483545, name="pos", field_type=LongType(), required=False
        ),
        schema_id=0,
    )

    partition_field_identity = PartitionField(
        source_id=1,
        field_id=101,
        transform=IdentityTransform(),
        name="number_partitioned",
    )
    partition_spec = PartitionSpec(partition_field_identity)

    properties = dict()
    properties["write.format.default"] = "parquet"
    properties["write.delete.mode"] = "merge-on-read"
    properties["write.update.mode"] = "merge-on-read"
    properties["write.merge.mode"] = "merge-on-read"
    properties["format-version"] = "2"

    drop_table_if_exists(identifier, session_catalog)
    session_catalog.create_table(
        identifier,
        schema=schema,
        partition_spec=partition_spec,
        properties=properties,
    )

    # 2. Use Spark to generate initial data files
    tbl = session_catalog.load_table(identifier)
    tbl.refresh()
    run_spark_commands(
        spark,
        [
            f"""
            INSERT INTO {identifier} VALUES (0, "pk1", "path1", 1), (0, "pk2", "path2", 2), (0, "pk3", "path3", 3)
            """
        ],
    )
    run_spark_commands(
        spark,
        [
            f"""
                INSERT INTO {identifier} VALUES (0, "pk1", "path1", 4), (0, "pk2", "path2", 5), (0, "pk3", "path3", 6)
                """
        ],
    )
    run_spark_commands(
        spark,
        [
            f"""
                INSERT INTO {identifier} VALUES (0, "pk4", "path4", 7), (0, "pk2", "path2", 8), (0, "pk3", "path3", 9)
                """
        ],
    )

    tbl = session_catalog.load_table(identifier)
    # 3. Use convert.remote() function to compute position deletes
    data_file_dict, equality_delete_dict, pos_delete_dict = fetch_all_bucket_files(tbl)

    convert_input_files_for_all_buckets = group_all_files_to_each_bucket(
        data_file_dict=data_file_dict,
        equality_delete_dict=equality_delete_dict,
        pos_delete_dict=pos_delete_dict,
    )

    s3_file_system = get_s3_file_system()
    for i, one_bucket_files in enumerate(convert_input_files_for_all_buckets):
        convert_input = ConvertInput.of(
            convert_input_files=one_bucket_files,
            convert_task_index=i,
            iceberg_table_warehouse_prefix="warehouse/default",
            identifier_fields=["primary_key"],
            compact_small_files=False,
            enforce_primary_key_uniqueness=True,
            position_delete_for_multiple_data_files=True,
            max_parallel_data_file_download=10,
            s3_file_system=s3_file_system,
        )

    number_partitioned_array_1 = pa.array([0, 0, 0], type=pa.int32())
    primary_key_array_1 = pa.array(["pk1", "pk2", "pk3"])
    names = ["number_partitioned", "primary_key"]
    data_table_1 = pa.Table.from_arrays(
        [number_partitioned_array_1, primary_key_array_1], names=names
    )

    number_partitioned_array_2 = pa.array([0, 0, 0], type=pa.int32())
    primary_key_array_2 = pa.array(["pk1", "pk2", "pk3"])
    names = ["number_partitioned", "primary_key"]
    data_table_2 = pa.Table.from_arrays(
        [number_partitioned_array_2, primary_key_array_2], names=names
    )

    number_partitioned_array_3 = pa.array([0, 0, 0], type=pa.int32())
    primary_key_array_3 = pa.array(["pk4", "pk2", "pk3"])
    names = ["number_partitioned", "primary_key"]
    data_table_3 = pa.Table.from_arrays(
        [number_partitioned_array_3, primary_key_array_3], names=names
    )

    download_data_mock = mocker.patch(
        "deltacat.compute.converter.utils.io.download_parquet_with_daft_hash_applied"
    )
    download_data_mock.side_effect = (data_table_1, data_table_2, data_table_3)

    convert_ref = convert.remote(convert_input)

    to_be_deleted_files_list = []
    to_be_added_files_dict_list = []
    convert_result = ray.get(convert_ref)

    partition_value = convert_input.convert_input_files.partition_value

    if convert_result[0]:
        to_be_deleted_files_list.extend(convert_result[0].values())

    file_location = convert_result[1][partition_value][0]
    to_be_added_files = f"s3://{file_location}"

    to_be_added_files_dict = defaultdict()
    to_be_added_files_dict[partition_value] = [to_be_added_files]
    to_be_added_files_dict_list.append(to_be_added_files_dict)

    # 4. Commit position delete, delete equality deletes from table
    new_position_delete_files = parquet_files_dict_to_iceberg_data_files(
        io=tbl.io,
        table_metadata=tbl.metadata,
        files_dict_list=to_be_added_files_dict_list,
    )
    commit_append_snapshot(
        iceberg_table=tbl,
        new_position_delete_files=new_position_delete_files,
    )
    tbl.refresh()

    # 5. Only primary key 2 and 3 should exist in table, as primary key 1 is deleted.
    pyiceberg_scan_table_rows = tbl.scan().to_arrow().to_pydict()

    # Only one unique record for each pk exists
    all_pk = sorted(pyiceberg_scan_table_rows["primary_key"])
    assert all_pk == ["pk1", "pk2", "pk3", "pk4"]

    # Expected unique record to keep for each pk
    expected_pk_to_pos_mapping = {"pk1": 4, "pk2": 8, "pk3": 9, "pk4": 7}
    for pk, pos in zip(
        pyiceberg_scan_table_rows["primary_key"], pyiceberg_scan_table_rows["pos"]
    ):
        assert pos == expected_pk_to_pos_mapping[pk]


@pytest.mark.integration
def test_converter_pos_delete_read_by_spark_success(
    spark, session_catalog: RestCatalog, setup_ray_cluster, mocker
) -> None:
    """
    Test for convert compute remote function happy case. Download file results are mocked.
    """

    # 1. Create Iceberg table
    namespace = "default"
    table_name = "table_converter_ray_pos_delete_read_by_spark_success"
    identifier = f"{namespace}.{table_name}"

    schema = Schema(
        NestedField(
            field_id=1, name="number_partitioned", field_type=LongType(), required=False
        ),
        NestedField(
            field_id=2, name="primary_key", field_type=StringType(), required=False
        ),
        schema_id=0,
    )

    partition_field_identity = PartitionField(
        source_id=1,
        field_id=101,
        transform=IdentityTransform(),
        name="number_partitioned",
    )
    partition_spec = PartitionSpec(partition_field_identity)

    properties = dict()
    properties["write.format.default"] = "parquet"
    properties["write.delete.mode"] = "merge-on-read"
    properties["write.update.mode"] = "merge-on-read"
    properties["write.merge.mode"] = "merge-on-read"
    properties["format-version"] = "2"

    drop_table_if_exists(identifier, session_catalog)
    session_catalog.create_table(
        identifier,
        schema=schema,
        partition_spec=partition_spec,
        properties=properties,
    )

    # 2. Use Spark to generate initial data files
    tbl = session_catalog.load_table(identifier)

    run_spark_commands(
        spark,
        [
            f"""
               INSERT INTO {identifier} VALUES (0, "pk1"), (0, "pk2"), (0, "pk3")
               """
        ],
    )
    run_spark_commands(
        spark,
        [
            f"""
                   INSERT INTO {identifier} VALUES (0, "pk1"), (0, "pk2"), (0, "pk3")
                   """
        ],
    )
    run_spark_commands(
        spark,
        [
            f"""
                   INSERT INTO {identifier} VALUES (0, "pk4"), (0, "pk2"), (0, "pk3")
                   """
        ],
    )
    tbl.refresh()

    # 3. Use convert.remote() function to compute position deletes
    data_file_dict, equality_delete_dict, pos_delete_dict = fetch_all_bucket_files(tbl)

    convert_input_files_for_all_buckets = group_all_files_to_each_bucket(
        data_file_dict=data_file_dict,
        equality_delete_dict=equality_delete_dict,
        pos_delete_dict=pos_delete_dict,
    )

    s3_file_system = get_s3_file_system()
    for i, one_bucket_files in enumerate(convert_input_files_for_all_buckets):
        convert_input = ConvertInput.of(
            convert_input_files=one_bucket_files,
            convert_task_index=i,
            iceberg_table_warehouse_prefix="warehouse/default",
            identifier_fields=["primary_key"],
            compact_small_files=False,
            enforce_primary_key_uniqueness=True,
            position_delete_for_multiple_data_files=True,
            max_parallel_data_file_download=10,
            s3_file_system=s3_file_system,
        )

    primary_key_array_1 = pa.array(["pk1", "pk2", "pk3"])
    names = ["primary_key"]
    data_table_1 = pa.Table.from_arrays([primary_key_array_1], names=names)

    primary_key_array_2 = pa.array(["pk1", "pk2", "pk3"])
    names = ["primary_key"]
    data_table_2 = pa.Table.from_arrays([primary_key_array_2], names=names)

    primary_key_array_3 = pa.array(["pk4", "pk2", "pk3"])
    names = ["primary_key"]
    data_table_3 = pa.Table.from_arrays([primary_key_array_3], names=names)

    download_data_mock = mocker.patch(
        "deltacat.compute.converter.utils.io.download_parquet_with_daft_hash_applied"
    )
    download_data_mock.side_effect = (data_table_1, data_table_2, data_table_3)

    convert_ref = convert.remote(convert_input)

    to_be_deleted_files_list = []
    to_be_added_files_dict_list = []
    convert_result = ray.get(convert_ref)

    partition_value = convert_input.convert_input_files.partition_value

    if convert_result[0]:
        to_be_deleted_files_list.extend(convert_result[0].values())

    file_location = convert_result[1][partition_value][0]
    to_be_added_files = f"s3://{file_location}"

    to_be_added_files_dict = defaultdict()
    to_be_added_files_dict[partition_value] = [to_be_added_files]
    to_be_added_files_dict_list.append(to_be_added_files_dict)

    # 4. Commit position delete, delete equality deletes from table
    new_position_delete_files = parquet_files_dict_to_iceberg_data_files(
        io=tbl.io,
        table_metadata=tbl.metadata,
        files_dict_list=to_be_added_files_dict_list,
    )

    commit_append_snapshot(
        iceberg_table=tbl,
        new_position_delete_files=new_position_delete_files,
    )
    tbl.refresh()

    # 5. Result assertion: Spark read table contains unique primary key
    spark_read_pos_delete = spark.sql(f"select * from {identifier}").collect()
    all_pk = [
        spark_read_pos_delete[row_idx][1]
        for row_idx in range(len(spark_read_pos_delete))
    ]
    all_pk_sorted = sorted(all_pk)
    assert all_pk_sorted == ["pk1", "pk2", "pk3", "pk4"]



from __future__ import annotations

import threading
from threading import Thread
from typing import Any, List, Set, Protocol, TypeVar, Dict, Iterable

from pyarrow import RecordBatch, Table
from deltacat.storage.model.partition import PartitionLocator
from deltacat.storage.rivulet.metastore.delta import ManifestIO, DeltacatManifestIO

from deltacat.storage.rivulet import Schema
from deltacat.storage.rivulet.metastore.json_sst import JsonSstWriter
from deltacat.storage.rivulet.serializer import MEMTABLE_DATA, DataSerializer
from deltacat.storage.rivulet.serializer_factory import DataSerializerFactory
from deltacat.storage.rivulet.writer.dataset_writer import DatasetWriter, DATA
from deltacat.storage.rivulet.metastore.sst import SSTWriter
from deltacat.storage.rivulet.fs.file_provider import FileProvider

INPUT_ROW = TypeVar("INPUT_ROW")


class Memtable(Protocol[INPUT_ROW]):
    """
    Protocol defining the interface for a memtable that can store and sort records of type T.
    """

    def add_record(self, record: INPUT_ROW) -> bool:
        """
        Add a record to the memtable.

        Args:
            record: The record to add of type INPUT_ROW

        Returns:
            bool: True if the memtable is full after adding the record, False otherwise
        """
        ...

    def get_sorted_records(self, schema: Schema) -> MEMTABLE_DATA:
        """
        Get all records in the memtable in sorted order.

        Returns:
            List[T]: A list of sorted records
        """
        ...


class DictMemTable(Memtable[Dict[str, Any]]):
    """
    Unit of in memory buffering of sorted records before records are written to file

    TODO future improvements:
        1. build b+ tree of record indexes on insertion
             OR If we end up using arrow as intermediate format, we can use
          pyarrow compute sort
        2. Probably we will re-write in rust
    """

    def __init__(self, merge_key: str):
        self.row_size = 0
        self.merge_key = merge_key

        self._records: List[Dict[str, Any]] = []
        self.lock = threading.Lock()

    def add_record(self, record: Dict[str, Any]):
        with self.lock:
            self._records.append(record)
            self.row_size += 1

        if self.row_size >= MemtableDatasetWriter.MAX_ROW_SIZE:
            return True
        return False

    def get_sorted_records(self, schema: Schema) -> List[Dict[str, Any]]:
        """
        Gets sorted records

        :return: iterator over sorted record
        """
        with self.lock:
            self._records.sort(key=lambda x: x.__getitem__(self.merge_key))
            return self._records


class RecordBatchMemTable(Memtable[RecordBatch]):
    """
    Note that this will not respect max row size.
    """

    def __init__(self, merge_key: str):
        self.row_size = 0
        self.merge_key = merge_key

        # list of full record batches in memtable
        self._records_batches: List[RecordBatch] = []
        self.lock = threading.Lock()

    def add_record(self, record: RecordBatch):
        with self.lock:
            self._records_batches.append(record)
            self.row_size += record.num_rows

        if self.row_size >= MemtableDatasetWriter.MAX_ROW_SIZE:
            return True
        return False

    def get_sorted_records(self, schema: Schema) -> Table:
        """
        Gets sorted records

        :return: iterator over sorted record
        """
        with self.lock:
            # Note that we are providing schema so that pyarrow does not infer it
            table = Table.from_batches(self._records_batches, schema.to_pyarrow())
            return table.sort_by(self.merge_key)


class MemtableDatasetWriter(DatasetWriter):
    # Note that this max row size is not respected when PyArrow RecordBatches are used
    # In that case, the entire record batch is written within one memtable even if the row count overflows
    MAX_ROW_SIZE = 1000000
    """
    Buffers data into rotating memtables. When a memtable reaches a certain size, it is flushed to disk and a new memtable is allocated

    Uses DataWriter which will be format specific for writing data
    Uses MetadataWriter for writing metadata

    TODO Future Improvements
    1. Maybe we should re-write this class in Rust (pending testing)
    """

    def __init__(
        self,
        file_provider: FileProvider,
        schema: Schema,
        locator: PartitionLocator,
        file_format: str | None = None,
        sst_writer: SSTWriter = None,
        manifest_io: ManifestIO = None,
    ):

        if not sst_writer:
            sst_writer = JsonSstWriter()
        if not manifest_io:
            manifest_io = DeltacatManifestIO(file_provider.uri, locator)

        self.schema = schema

        self.file_provider = file_provider
        self.data_serializer: DataSerializer = DataSerializerFactory.get_serializer(
            self.schema, self.file_provider, file_format
        )
        self.sst_writer = sst_writer
        self.manifest_io = manifest_io

        self._sst_files: Set[str] = set()
        self.__curr_memtable = None
        self.__open_memtables = []
        self.__rlock = threading.RLock()
        self.__open_threads: List[Thread] = []
        self._locator = locator

    def write_dict(self, record: Dict[str, Any]) -> None:

        # Construct memtable if doesn't exist. If previous memtable wrong type, rotate
        memtable_ctor = lambda: DictMemTable(self.schema.get_merge_key())
        if not self.__curr_memtable:
            self.__curr_memtable = memtable_ctor()
        try:
            isinstance(self.__curr_memtable, DictMemTable)
        except TypeError:
            self.__rotate_memtable(memtable_ctor)

        # Write record(s). If memtable is full, rotate
        if self.__curr_memtable.add_record(record):
            self.__rotate_memtable(memtable_ctor)

    def write_record_batch(self, record: RecordBatch) -> None:
        # Construct memtable if doesn't exist. If previous memtable wrong type, rotate
        memtable_ctor = lambda: RecordBatchMemTable(self.schema.get_merge_key())
        if not self.__curr_memtable:
            self.__curr_memtable = memtable_ctor()

        try:
            isinstance(self.__curr_memtable, RecordBatchMemTable)
        except TypeError:
            self.__rotate_memtable(memtable_ctor)

        # Write record(s). If memtable is full, rotate
        if self.__curr_memtable.add_record(record):
            self.__rotate_memtable(memtable_ctor)

    def write(self, data: DATA) -> None:
        if isinstance(data, RecordBatch):
            self.write_record_batch(data)
        elif isinstance(data, Iterable):
            for x in data:
                if isinstance(x, dict):
                    self.write_dict(x)
                elif isinstance(x, RecordBatch):
                    self.write_record_batch(x)
                else:
                    raise ValueError(
                        f"Iterable contained unsupported type {type(x).__name__}."
                        f" Supported data types to write are: {DATA}"
                    )
        else:
            raise ValueError(
                f"Unsupported data type {type(data).__name__}. Supported data types to write are: {DATA}"
            )

    def flush(self) -> str:
        """
        Explicitly flush any data and metadata and commit to dataset
        """
        self.__flush_memtable(self.__curr_memtable)
        for thread in [t for t in self.__open_threads if t.is_alive()]:
            thread.join()

        manifest_location = self.__write_manifest_file()
        self._sst_files.clear()

        return manifest_location

    def __enter__(self) -> Any:
        """
        Enter and exit method allows python "with" statement
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes all open memtables and ensures all data is flushed.
        """
        self.flush()
        # return False to propogate up error messages
        return False

    def __rotate_memtable(self, memtable_constructor_closure):
        """
        Replace the active memtable
        :return:
        """
        with self.__rlock:
            self.__flush_memtable(self.__curr_memtable)
            self.__curr_memtable = memtable_constructor_closure()
            self.__open_memtables.append(self.__curr_memtable)

            # Reap dead threads
            self.__open_threads = [t for t in self.__open_threads if t.is_alive()]

    def __flush_memtable(self, memtable):
        thread = threading.Thread(target=self.__flush_memtable_async, args=(memtable,))
        thread.start()
        with self.__rlock:
            self.__open_threads.append(thread)

    def __flush_memtable_async(self, memtable: Memtable):
        """
        Flushes data and metadata for a given memtable
        Called asynchronously in background thread
        """
        if not memtable:
            return

        sst_metadata_list = self.data_serializer.flush_batch(
            memtable.get_sorted_records(self.schema)
        )

        # short circuit if no data/metadata written
        if not sst_metadata_list:
            with self.__rlock:
                self.__open_memtables.remove(memtable)
            return

        # Write SST. Each memtable is going to have a dedicated L0 SST file because that is the unit at which
        # we have contiguously sorted data
        sst_file = self.file_provider.provide_l0_sst_file()

        with self.__rlock:
            self.sst_writer.write(sst_file, sst_metadata_list)
            self._sst_files.add(sst_file.location)

            if memtable in self.__open_memtables:
                self.__open_memtables.remove(memtable)

    def __write_manifest_file(self) -> str:
        """
        Write the manifest file to the filesystem at the given URI.
        """
        return self.manifest_io.write(list(self._sst_files), self.schema, 0)



# Allow self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
from types import ModuleType

from typing import Any, Dict, List, Optional, Union
from functools import partial
import ray

from deltacat import logs
from deltacat.annotations import ExperimentalAPI
from deltacat.catalog.main import impl as DeltacatCatalog
from deltacat.catalog.iceberg import impl as IcebergCatalog
from deltacat.catalog import CatalogProperties
from deltacat.catalog.iceberg import IcebergCatalogConfig
from deltacat.constants import DEFAULT_CATALOG

all_catalogs: Optional[ray.actor.ActorHandle] = None

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class Catalog:
    def __init__(self, impl: ModuleType = DeltacatCatalog, *args, **kwargs):
        """
        Constructor for a Catalog.

        The args and kwargs here will be plumbed through to the catalog initialize function, and the results
        are stored in Catalog.inner. Any state which is required (like: metastore root URI, pyiceberg native catalog)
        MUST be returned by initialize.

        Note: all initialization configuration MUST be pickle-able. When `Catalog` is pickled, _inner is excluded.
        Instead, we only pass impl/args/kwargs, which are pickled and then _inner is re-constituted by calling __init__.
        See `ray.util.register_serializer` in Catalogs class.
        """
        if not isinstance(self, Catalog):
            # self may contain the tuple returned from __reduce__ (ray pickle bug?)
            if callable(self[0]) and isinstance(self[1], tuple):
                logger.info(f"Invoking {self[0]} with positional args: {self[1]}")
                return self[0](*self[1])
            else:
                err_msg = f"Expected `self` to be {Catalog}, but found: {self}"
                raise RuntimeError(err_msg)

        self._impl = impl
        self._inner = self._impl.initialize(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs

    @classmethod
    @ExperimentalAPI
    def iceberg(cls, config: IcebergCatalogConfig, *args, **kwargs):
        """
        !!! ICEBERG SUPPORT IS EXPERIMENTAL !!!

        Factory method to construct a catalog from Iceberg catalog params

        This method is just a wrapper around __init__ with stronger typing. You may still call __init__,
        plumbing __params__ through as kwargs
        """
        return cls(impl=IcebergCatalog, *args, **{"config": config, **kwargs})

    @classmethod
    def default(cls, config: CatalogProperties, *args, **kwargs):
        """
        Factory method to construct a catalog with the default implementation

        Uses CatalogProperties as configuration
        """
        return cls(impl=DeltacatCatalog, *args, **{"config": config, **kwargs})

    @property
    def impl(self):
        return self._impl

    @property
    def inner(self) -> Optional[Any]:
        return self._inner

    # support pickle, copy, deepcopy, etc.
    def __reduce__(self):
        # instantiated catalogs may fail to pickle, so exclude _inner
        # (e.g. Iceberg catalog w/ unserializable SSLContext from boto3 client)
        return partial(self.__class__, **self._kwargs), (self._impl, *self._args)

    def __str__(self):
        string_rep = f"{self.__class__.__name__}("
        if self._args:
            string_rep += f"args={self._args}, "
        if self._kwargs:
            string_rep += f"kwargs={self._kwargs}, "
        if self._inner:
            string_rep += f"inner={self._inner})"
        return string_rep

    def __repr__(self):
        return self.__str__()


@ray.remote
class Catalogs:
    def __init__(
        self,
        catalogs: Union[Catalog, Dict[str, Catalog]],
        default: Optional[str] = None,
        *args,
        **kwargs,
    ):
        if default and default not in catalogs:
            raise ValueError(
                f"Catalog {default} not found " f"in catalogs to register: {catalogs}"
            )
        if not catalogs:
            raise ValueError(
                f"No catalogs given to register. "
                f"Please specify one or more catalogs."
            )

        # if user only provides single Catalog, override it to be a map with default key
        if isinstance(catalogs, Catalog):
            catalogs = {DEFAULT_CATALOG: catalogs}

        self.catalogs: Dict[str, Catalog] = catalogs
        if default:
            self.default_catalog = self.catalogs[default]
        elif len(catalogs) == 1:
            self.default_catalog = list(self.catalogs.values())[0]
        else:
            self.default_catalog = None

    def all(self) -> Dict[str, Catalog]:
        return self.catalogs

    def names(self) -> List[str]:
        return list(self.catalogs.keys())

    def put(self, name: str, catalog: Catalog, set_default: bool = False) -> None:
        self.catalogs[name] = catalog
        if set_default:
            self.default_catalog = catalog

    def get(self, name) -> Catalog:
        return self.catalogs.get(name)

    def default(self) -> Optional[Catalog]:
        return self.default_catalog


def is_initialized(*args, **kwargs) -> bool:
    """
    Check if DeltaCAT is initialized
    """
    global all_catalogs

    # If ray is not initialized, then Catalogs cannot be initialized
    if not ray.is_initialized():
        # Any existing actor reference stored in catalog_module must be stale - reset it
        all_catalogs = None
        return False

    return all_catalogs is not None


def init(
    catalogs: Union[Dict[str, Catalog], Catalog],
    default: Optional[str] = None,
    ray_init_args: Dict[str, Any] = None,
    *args,
    force_reinitialize=False,
    **kwargs,
) -> None:
    """
    Initialize DeltaCAT catalogs.

    :param catalogs: Either a single Catalog instance or a map of string to Catalog instance
    :param default: The Catalog to use by default. If only one Catalog is provided, it will
        be set as the default
    :param ray_init_args: kwargs to pass to ray initialization
    :param force_reinitialize: if True, force the reinitialization of Ray. If false, will do nothing if ray already initialized
    """
    global all_catalogs

    if is_initialized() and not force_reinitialize:
        logger.warning("DeltaCAT already initialized.")
        return
    else:
        if ray_init_args:
            ray.init(**ray_init_args)
        else:
            ray.init()

    # register custom serializer for catalogs since these may contain
    # unserializable objects like boto3 clients with SSLContext
    ray.util.register_serializer(
        Catalog, serializer=Catalog.__reduce__, deserializer=Catalog.__init__
    )
    all_catalogs = Catalogs.remote(catalogs=catalogs, default=default)


def get_catalog(name: Optional[str] = None, **kwargs) -> Catalog:
    """
    Get a catalog by name, or the default catalog if no name is provided.

    Args:
        name: Name of catalog to retrieve (optional, uses default if not provided)

    Returns:
        The requested Catalog, or ValueError if it does not exist
    """
    global all_catalogs

    if not all_catalogs:
        raise ValueError(
            "No catalogs available! Call "
            "`deltacat.init(catalogs={...})` to register one or more "
            "catalogs then retry."
        )

    if name is not None:
        catalog = ray.get(all_catalogs.get.remote(name))
        if not catalog:
            available_catalogs = ray.get(all_catalogs.all.remote()).values()
            raise ValueError(
                f"Catalog '{name}' not found. Available catalogs: "
                f"{available_catalogs}."
            )
        return catalog

    else:
        catalog = ray.get(all_catalogs.default.remote())
        if not catalog:
            available_catalogs = ray.get(all_catalogs.all.remote()).values()
            raise ValueError(
                f"Call to get_catalog without name set failed because there is no default Catalog set. Available catalogs: "
                f"{available_catalogs}."
            )
        return catalog


def put_catalog(
    name: str,
    catalog: Catalog = None,
    *,
    default: bool = False,
    ray_init_args: Dict[str, Any] = None,
    fail_if_exists: bool = False,
    **kwargs,
) -> None:
    """
    Add a named catalog to the global map of named catalogs. Initializes ray if not already initialized.

    Args:
        name: name of catalog
        catalog: catalog instance to use, if provided
        default:  Make this the default catalog if multiple catalogs are available.
            ignored if this is the only catalog available, since it will always be the default catalog.
        ray_init_args: ray initialization args (used only if ray not already initialized)
        fail_if_exists: if True, raises KeyError if the catalog name already exists. Otherwise, overwrite catalog
    """
    global all_catalogs

    # Initialize, if necessary
    if not is_initialized():
        # NOTE - since we are initializing with a single catalog, it will be set to the default
        if not default:
            logger.info(
                f"Calling put_catalog with set_as_default=False, "
                f"but still setting Catalog {catalog} as default since it is the only catalog."
            )
        init({name: catalog}, ray_init_args=ray_init_args)
        return

    # Fail if fail_if_exists and catalog already exists
    if fail_if_exists:
        catalog_already_exists = False
        try:
            get_catalog(name)
            # Note - need to set state catalog_already_exists and throw ValueError later, or else it will be
            # caught in the except block which is meant to catch the ValueError from get_catalog
            catalog_already_exists = True
        except ValueError:
            pass
        if catalog_already_exists:
            raise ValueError(
                f"Failed to put catalog {name} because it already exists and fail_if_exists={fail_if_exists}"
            )

    # Add the catalog (which may overwrite existing if fail_if_exists=False)
    ray.get(all_catalogs.put.remote(name, catalog, default))



import os
import tempfile
import shutil
from typing import Dict

import pytest


@pytest.fixture
def temp_dir():
    """
    Fixture that creates a temporary directory for tests and cleans it up afterwards.

    Returns:
        str: Path to the temporary directory
    """
    # Create a temporary directory
    dir_path = tempfile.mkdtemp()

    # Provide the directory path to the test
    yield dir_path

    # Cleanup: remove the directory after the test is done
    shutil.rmtree(dir_path)


@pytest.fixture(scope="function")
def local_deltacat_storage_kwargs(temp_dir):
    """
    Fixture that creates a temporary database file for each test function
    and returns storage kwargs dictionary.

    Returns:
        dict: A dictionary with db_file_path key pointing to a temporary database file
    """
    # Create a unique database file in the temporary directory
    db_file_path = os.path.join(temp_dir, "db_test.sqlite")

    # Return kwargs dictionary ready to use
    kwargs = {"db_file_path": db_file_path}
    yield kwargs

    # Cleanup: remove the database file if it exists
    if os.path.exists(db_file_path):
        os.remove(db_file_path)


def create_local_deltacat_storage_file() -> Dict[str, str]:
    """
    Helper function to create a local deltacat storage file

    Essentially uses the same approach as local_deltacat_storage_kwargs, but more flexible
    if the consumer does not want to use a function scoped fixture

    Returns: kwargs to use for local deltacat storage, i.e. {"db_file_path": $db_file}
    """
    temp_dir = tempfile.mkdtemp()
    db_file_path = os.path.join(temp_dir, "db_test.sqlite")
    return {"db_file_path": db_file_path}


def clean_up_local_deltacat_storage_file(local_storage_kwargs: Dict[str, str]):
    """
    Cleans up local file and directory created by create_local_deltacat_storage_file
    """
    db_file = local_storage_kwargs["db_file_path"]
    dir_path = os.path.dirname(db_file)

    # Remove the database file if it exists
    if os.path.exists(db_file):
        os.remove(db_file)

    # Remove the temporary directory if it exists
    if os.path.exists(dir_path):
        shutil.rmtree(dir_path)



import logging

from typing import Any, Dict, List, Optional, Union

from daft import DataFrame, context
from daft.daft import ScanOperatorHandle, StorageConfig
from daft.logical.builder import LogicalPlanBuilder

from deltacat import logs
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.daft.daft_scan import DeltaCatScanOperator
from deltacat.exceptions import TableAlreadyExistsError
from deltacat.storage.iceberg.iceberg_scan_planner import IcebergScanPlanner
from deltacat.storage.iceberg.model import PartitionSchemeMapper, SchemaMapper
from deltacat.storage.model.partition import PartitionScheme
from deltacat.storage.iceberg.impl import _get_native_catalog
from deltacat.storage.model.sort_key import SortScheme
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace, NamespaceProperties
from deltacat.storage.model.schema import Schema
from deltacat.storage.model.table import TableProperties
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    StreamFormat,
)
from deltacat.storage.iceberg import impl as IcebergStorage
from deltacat.types.media import ContentType
from deltacat.types.tables import TableWriteMode
from deltacat.constants import DEFAULT_NAMESPACE
from deltacat.catalog.iceberg.iceberg_catalog_config import IcebergCatalogConfig

from pyiceberg.catalog import Catalog, load_catalog
from pyiceberg.transforms import BucketTransform

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


# catalog functions
def initialize(*args, config: IcebergCatalogConfig, **kwargs) -> Catalog:
    """
    Initializes an Iceberg catalog with the given config.

    NOTE: because PyIceberg catalogs are not pickle-able, we cannot accept them as catalog initialization parameters,
    since catalog initialization parameters are passed to Ray actors (see: :class:`deltacat.catalog.Catalogs`)

    Args:
        **kwargs: Arguments to be passed to PyIceberg Catalog.
            If 'catalog' is provided as a PyIceberg Catalog instance, it will be used directly.
            Otherwise, the arguments will be used to load a catalog via pyiceberg.catalog.load_catalog.

    Returns:
        IcebergCatalogConfig: Configuration wrapper containing the PyIceberg Catalog.
    """

    # If no catalog is provided, try to load one with PyIceberg

    load_catalog_args = {"type": config.type.value, **config.properties, **kwargs}
    catalog = load_catalog(**load_catalog_args)
    return catalog


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],
    table: str,
    *args,
    namespace: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    **kwargs,
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments."""

    # TODO (pdames): derive schema automatically from data if not
    #  explicitly specified in kwargs, and table needs to be created
    # kwargs["schema"] = kwargs["schema"] or derived_schema
    kwargs["fail_if_exists"] = mode == TableWriteMode.CREATE
    table_definition = (
        create_table(
            table,
            namespace=namespace,
            *args,
            **kwargs,
        )
        if (mode == TableWriteMode.AUTO or mode == TableWriteMode.CREATE)
        else get_table(table, namespace=namespace, *args, **kwargs)
    )

    # TODO(pdames): Use native DeltaCAT models to map from Iceberg partitioning to Daft partitioning...
    #  this lets us re-use a single model-mapper instead of different per-catalog model mappers
    schema = SchemaMapper.unmap(table_definition.table_version.schema)
    partition_spec = PartitionSchemeMapper.unmap(
        table_definition.table_version.partition_scheme,
        schema,
    )
    if isinstance(data, DataFrame):
        for partition_field in partition_spec.fields:
            if isinstance(partition_field.transform, BucketTransform):
                ice_bucket_transform: BucketTransform = partition_field.transform
                # TODO(pdames): Get a type-checked Iceberg Table automatically via unmap()
                table_location = table_definition.table.native_object.location()
                path = kwargs.get("path") or f"{table_location}/data"
                if content_type == ContentType.PARQUET:
                    source_field = schema.find_field(
                        name_or_id=partition_field.source_id
                    )
                    out_df = data.write_parquet(
                        path,
                        partition_cols=[
                            data[source_field.name].partitioning.iceberg_bucket(
                                ice_bucket_transform.num_buckets
                            ),
                        ],
                    )
                    # TODO(pdames): only append s3:// to output file paths when writing to S3!
                    out_file_paths = [f"s3://{val}" for val in out_df.to_arrow()[0]]
                    from deltacat.catalog.iceberg import overrides

                    overrides.append(
                        table_definition.table.native_object,
                        out_file_paths,
                    )
                else:
                    raise NotImplementedError(
                        f"iceberg writes not implemented for content type: {content_type}"
                    )
            else:
                raise NotImplementedError(
                    f"daft partitioning not implemented for iceberg transform: {partition_field.transform}"
                )
    else:
        raise NotImplementedError(
            f"iceberg write-back not implemented for data type: {type(data)}"
        )


def read_table(
    table: str, *args, namespace: Optional[str] = None, **kwargs
) -> DistributedDataset:
    """Read a table into a distributed dataset."""
    # TODO: more proper IO configuration
    io_config = context.get_context().daft_planning_config.default_io_config
    multithreaded_io = context.get_context().get_or_create_runner().name != "ray"

    storage_config = StorageConfig(multithreaded_io, io_config)

    dc_table = get_table(name=table, namespace=namespace, **kwargs)
    dc_scan_operator = DeltaCatScanOperator(dc_table, storage_config)
    handle = ScanOperatorHandle.from_python_scan_operator(dc_scan_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)


def alter_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    **kwargs,
) -> None:
    """Alter table definition."""
    raise NotImplementedError("alter_table not implemented")


def create_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    namespace_properties: Optional[NamespaceProperties] = None,
    content_types: Optional[List[ContentType]] = None,
    fail_if_exists: bool = True,
    **kwargs,
) -> TableDefinition:
    """Create an empty table in the catalog"""

    namespace = namespace or default_namespace()
    existing_table = get_table(
        name,
        *args,
        namespace=namespace,
        **kwargs,
    )
    if existing_table:
        if fail_if_exists:
            err_msg = (
                f"Table `{namespace}.{name}` already exists. "
                f"To suppress this error, rerun `create_table()` with "
                f"`fail_if_exists=False`."
            )
            raise TableAlreadyExistsError(err_msg)
        else:
            logger.debug(f"Returning existing table: `{namespace}.{name}`")
            return existing_table

    if not IcebergStorage.namespace_exists(namespace, **kwargs):
        logger.debug(f"Namespace {namespace} doesn't exist. Creating it...")
        IcebergStorage.create_namespace(
            namespace,
            properties=namespace_properties or {},
            **kwargs,
        )

    IcebergStorage.create_table_version(
        namespace=namespace,
        table_name=name,
        table_version=version,
        schema=schema,
        partition_scheme=partition_scheme,
        sort_keys=sort_keys,
        table_properties=table_properties,
        **kwargs,
    )

    return get_table(
        name,
        *args,
        namespace=namespace,
        **kwargs,
    )


def drop_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    purge: bool = False,
    **kwargs,
) -> None:
    """Drop a table from the catalog and optionally purge it. Raises an error
    if the table does not exist."""
    raise NotImplementedError("drop_table not implemented")


def refresh_table(table: str, *args, namespace: Optional[str] = None, **kwargs) -> None:
    """Refresh metadata cached on the Ray cluster for the given table."""
    raise NotImplementedError("refresh_table not implemented")


def list_tables(
    *args, namespace: Optional[str] = None, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions. Raises an error if the given namespace
    does not exist."""
    raise NotImplementedError("list_tables not implemented")


def get_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    **kwargs,
) -> Optional[TableDefinition]:
    """Get table definition metadata.

    Args:
        name: Name of the table to retrieve
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to retrieve.
            If not specified, the latest version is used.
        stream_format: Optional stream format to retrieve

    Returns:
        Deltacat TableDefinition if the table exists, None otherwise.
    """
    namespace = namespace or default_namespace()
    stream = IcebergStorage.get_stream(namespace=namespace, table_name=name, **kwargs)
    if not stream:
        return None
    table_obj = IcebergStorage.get_table(namespace=namespace, table_name=name, **kwargs)
    if not table_obj:
        return None
    table_version_obj = None
    if table_version:
        table_version_obj = IcebergStorage.get_table_version(
            namespace=namespace, table_name=name, table_version=table_version, **kwargs
        )
    else:
        table_version_obj = IcebergStorage.get_latest_table_version(
            namespace=namespace, table_name=name, **kwargs
        )
    if not table_version_obj:
        return None
    scan_planner = IcebergScanPlanner(_get_native_catalog(**kwargs))
    return TableDefinition.of(
        table=table_obj,
        table_version=table_version_obj,
        stream=stream,
        native_object=table_obj.native_object,
        scan_planner=scan_planner,
    )


def truncate_table(
    table: str, *args, namespace: Optional[str] = None, **kwargs
) -> None:
    """Truncate table data. Raises an error if the table does not exist."""
    raise NotImplementedError("truncate_table not implemented")


def rename_table(
    table: str, new_name: str, *args, namespace: Optional[str] = None, **kwargs
) -> None:
    """Rename a table."""
    raise NotImplementedError("rename_table not implemented")


def table_exists(table: str, *args, namespace: Optional[str] = None, **kwargs) -> bool:
    """Returns True if the given table exists, False if not."""
    namespace = namespace or default_namespace()
    return IcebergStorage.table_exists(namespace=namespace, table_name=table, **kwargs)


# namespace functions
def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """List a page of table namespaces."""
    return IcebergStorage.list_namespaces(**kwargs)


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """Gets table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist."""
    return IcebergStorage.get_namespace(namespace, **kwargs)


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """Returns True if the given table namespace exists, False if not."""
    return IcebergStorage.namespace_exists(namespace, **kwargs)


def create_namespace(
    namespace: str, *args, properties: Optional[NamespaceProperties] = None, **kwargs
) -> Namespace:
    """Creates a table namespace with the given name and properties. Returns
    the created namespace. Raises an error if the namespace already exists."""
    raise NotImplementedError("create_namespace not implemented")


def alter_namespace(
    namespace: str,
    *args,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    **kwargs,
) -> None:
    """Alter table namespace definition."""
    raise NotImplementedError("alter_namespace not implemented")


def drop_namespace(namespace: str, *args, purge: bool = False, **kwargs) -> None:
    """Drop the given namespace and all of its tables from the catalog,
    optionally purging them."""
    raise NotImplementedError("drop_namespace not implemented")


def default_namespace(*args, **kwargs) -> str:
    """Returns the default namespace for the catalog."""
    return DEFAULT_NAMESPACE



from typing import Any, Dict, List, Optional, Union, Tuple
import logging

from deltacat.catalog import CatalogProperties
from deltacat.exceptions import (
    NamespaceAlreadyExistsError,
    StreamNotFoundError,
    TableAlreadyExistsError,
    TableVersionNotFoundError,
)
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.sort_key import SortScheme
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace, NamespaceProperties
from deltacat.storage.model.schema import Schema
from deltacat.storage.model.table import TableProperties, Table
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    StreamFormat,
)
from deltacat.storage.model.partition import (
    Partition,
    PartitionLocator,
    PartitionScheme,
)
from deltacat.storage.model.table_version import TableVersion
from deltacat.compute.merge_on_read.model.merge_on_read_params import MergeOnReadParams
from deltacat.storage.model.delta import DeltaType
from deltacat.types.media import ContentType, TableType, DistributedDatasetType
from deltacat.types.tables import TableWriteMode
from deltacat.compute.merge_on_read import MERGE_FUNC_BY_DISTRIBUTED_DATASET_TYPE
from deltacat import logs
from deltacat.constants import DEFAULT_NAMESPACE
from deltacat.storage import metastore as storage_impl

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

"""
This is the default implementation for the Catalog interface, using DeltaCAT native storage

Note that, when this catalog implementation gets called through the normal pattern of `delegate.py`, all functions
will be called the kwarg "inner" equal to the `CatalogProperties` this was initialized with.

`CatalogProperties` has all state required to implement catalog functions, such as metastore root URI
"""


# catalog functions
def initialize(config: CatalogProperties = None, *args, **kwargs) -> CatalogProperties:
    """
    Initializes the data catalog with the given arguments.

    returns CatalogProperties as the "inner" state value for a DC native catalog
    """
    if config is not None:
        return config
    else:
        return CatalogProperties(*args, **kwargs)


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],  # type: ignore
    table: str,
    *args,
    namespace: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    **kwargs,
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments."""
    raise NotImplementedError("write_to_table not implemented")


def read_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    table_type: Optional[TableType] = TableType.PYARROW,
    distributed_dataset_type: Optional[
        DistributedDatasetType
    ] = DistributedDatasetType.RAY_DATASET,
    partition_filter: Optional[List[Union[Partition, PartitionLocator]]] = None,
    stream_position_range_inclusive: Optional[Tuple[int, int]] = None,
    merge_on_read: Optional[bool] = False,
    reader_kwargs: Optional[Dict[Any, Any]] = None,
    **kwargs,
) -> DistributedDataset:  # type: ignore
    """Read a table into a distributed dataset."""

    if reader_kwargs is None:
        reader_kwargs = {}

    _validate_read_table_args(
        namespace=namespace,
        table_type=table_type,
        distributed_dataset_type=distributed_dataset_type,
        merge_on_read=merge_on_read,
        **kwargs,
    )

    table_version_obj = _get_latest_or_given_table_version(
        namespace=namespace,
        table_name=table,
        table_version=table_version,
        **kwargs,
    )
    table_version = table_version_obj.table_version

    if (
        table_version_obj.content_types is None
        or len(table_version_obj.content_types) != 1
    ):
        raise ValueError(
            "Expected exactly one content type but "
            f"found {table_version_obj.content_types}."
        )

    logger.info(
        f"Reading metadata for table={namespace}/{table}/{table_version} "
        f"with partition_filters={partition_filter} and stream position"
        f" range={stream_position_range_inclusive}"
    )

    if partition_filter is None:
        logger.info(
            f"Reading all partitions metadata in the table={table} "
            "as partition_filter was None."
        )
        partition_filter = (
            _get_storage(**kwargs)
            .list_partitions(
                table_name=table,
                namespace=namespace,
                table_version=table_version,
                **kwargs,
            )
            .all_items()
        )

    qualified_deltas = _get_deltas_from_partition_filter(
        stream_position_range_inclusive=stream_position_range_inclusive,
        partition_filter=partition_filter,
        **kwargs,
    )

    logger.info(
        f"Total qualified deltas={len(qualified_deltas)} "
        f"from {len(partition_filter)} partitions."
    )

    merge_on_read_params = MergeOnReadParams.of(
        {
            "deltas": qualified_deltas,
            "deltacat_storage": _get_storage(**kwargs),
            "deltacat_storage_kwargs": {**kwargs},
            "reader_kwargs": reader_kwargs,
        }
    )

    return MERGE_FUNC_BY_DISTRIBUTED_DATASET_TYPE[distributed_dataset_type.value](
        params=merge_on_read_params, **kwargs
    )


def alter_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    properties: Optional[TableProperties] = None,
    **kwargs,
) -> None:
    """Alter deltacat table/table_version definition.

    Modifies various aspects of a table's metadata including lifecycle state,
    schema, partitioning, sort keys, description, and properties.

    Args:
        table: Name of the table to alter.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        lifecycle_state: New lifecycle state for the table.
        schema_updates: Map of schema updates to apply.
        partition_updates: Map of partition scheme updates to apply.
        sort_keys: New sort keys scheme.
        description: New description for the table.
        properties: New table properties.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not already exist.
    """
    namespace = namespace or default_namespace()

    _get_storage(**kwargs).update_table(
        *args,
        namespace=namespace,
        table_name=table,
        description=description,
        properties=properties,
        lifecycle_state=lifecycle_state,
        **kwargs,
    )

    table_version = _get_storage(**kwargs).get_latest_table_version(
        namespace, table, **kwargs
    )
    _get_storage(**kwargs).update_table_version(
        *args,
        namespace=namespace,
        table_name=table,
        table_version=table_version.id,
        description=description,
        schema_updates=schema_updates,
        partition_updates=partition_updates,
        sort_keys=sort_keys,
        **kwargs,
    )


def create_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = LifecycleState.ACTIVE,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    namespace_properties: Optional[NamespaceProperties] = None,
    content_types: Optional[List[ContentType]] = None,
    fail_if_exists: bool = True,
    **kwargs,
) -> TableDefinition:
    """Create an empty table in the catalog.

    If a namespace isn't provided, the table will be created within the default deltacat namespace.
    Additionally if the provided namespace does not exist, it will be created for you.


    Args:
        name: Name of the table to create.
        namespace: Optional namespace for the table. Uses default namespace if not specified.
        version: Optional version identifier for the table.
        lifecycle_state: Lifecycle state of the new table. Defaults to ACTIVE.
        schema: Schema definition for the table.
        partition_scheme: Optional partitioning scheme for the table.
        sort_keys: Optional sort keys for the table.
        description: Optional description of the table.
        table_properties: Optional properties for the table.
        namespace_properties: Optional properties for the namespace if it needs to be created.
        content_types: Optional list of allowed content types for the table.
        fail_if_exists: If True, raises an error if table already exists. If False, returns existing table.

    Returns:
        TableDefinition object for the created or existing table.

    Raises:
        TableAlreadyExistsError: If the table already exists and fail_if_exists is True.
        NamespaceNotFoundError: If the provided namespace does not exist.
    """
    namespace = namespace or default_namespace()

    table = get_table(*args, name, namespace=namespace, table_version=version, **kwargs)
    if table is not None:
        if fail_if_exists:
            raise TableAlreadyExistsError(f"Table {namespace}.{name} already exists")
        return table

    if not namespace_exists(*args, namespace, **kwargs):
        create_namespace(
            *args, namespace=namespace, properties=namespace_properties, **kwargs
        )

    (table, table_version, stream) = _get_storage(**kwargs).create_table_version(
        *args,
        namespace=namespace,
        table_name=name,
        table_version=version,
        schema=schema,
        partition_scheme=partition_scheme,
        sort_keys=sort_keys,
        table_version_description=description,
        table_description=description,
        table_properties=table_properties,
        lifecycle_state=lifecycle_state or LifecycleState.ACTIVE,
        supported_content_types=content_types,
        **kwargs,
    )

    return TableDefinition.of(
        table=table,
        table_version=table_version,
        stream=stream,
    )


def drop_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    purge: bool = False,
    **kwargs,
) -> None:
    """Drop a table from the catalog and optionally purges underlying data.

    Args:
        name: Name of the table to drop.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        purge: If True, permanently delete the table data. If False, only remove from catalog.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not exist.

    TODO: Honor purge once garbage collection is implemented.
    TODO: Drop table version if specified, possibly create a delete_table_version api.
    """
    if purge:
        raise NotImplementedError("Purge flag is not currently supported.")

    namespace = namespace or default_namespace()
    _get_storage(**kwargs).delete_table(
        *args, namespace=namespace, name=name, purge=purge, **kwargs
    )


def refresh_table(table: str, *args, namespace: Optional[str] = None, **kwargs) -> None:
    """Refresh metadata cached on the Ray cluster for the given table.

    Args:
        table: Name of the table to refresh.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        None
    """
    raise NotImplementedError("refresh_table not implemented")


def list_tables(
    *args, namespace: Optional[str] = None, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions.

    Args:
        namespace: Optional namespace to list tables from. Uses default namespace if not specified.

    Returns:
        ListResult containing TableDefinition objects for tables in the namespace.
    """
    namespace = namespace or default_namespace()
    tables = _get_storage(**kwargs).list_tables(*args, namespace=namespace, **kwargs)
    table_definitions = [
        get_table(*args, table.table_name, namespace, **kwargs)
        for table in tables.all_items()
    ]

    return ListResult(items=table_definitions)


def get_table(
    name: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    **kwargs,
) -> Optional[TableDefinition]:
    """Get table definition metadata.

    Args:
        name: Name of the table to retrieve.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to retrieve.
            If not specified, the latest version is used.
        stream_format: Optional stream format to retrieve. Uses the default Deltacat stream
            format if not specified.

    Returns:
        Deltacat TableDefinition if the table exists, None otherwise.

    Raises:
        TableVersionNotFoundError: If the table version does not exist.
        StreamNotFoundError: If the stream does not exist.
    """
    namespace = namespace or default_namespace()
    table: Optional[Table] = _get_storage(**kwargs).get_table(
        *args, table_name=name, namespace=namespace, **kwargs
    )

    if table is None:
        return None

    table_version: Optional[TableVersion] = _get_storage(**kwargs).get_table_version(
        *args, namespace, name, table_version or table.latest_table_version, **kwargs
    )

    if table_version is None:
        raise TableVersionNotFoundError(
            f"TableVersion {namespace}.{name}.{table_version} does not exist."
        )

    stream = _get_storage(**kwargs).get_stream(
        *args,
        namespace=namespace,
        table_name=name,
        table_version=table_version.id,
        stream_format=stream_format,
        **kwargs,
    )

    if stream is None:
        raise StreamNotFoundError(
            f"Stream {namespace}.{table}.{table_version}.{stream} does not exist."
        )

    return TableDefinition.of(
        table=table,
        table_version=table_version,
        stream=stream,
    )


def truncate_table(
    table: str, *args, namespace: Optional[str] = None, **kwargs
) -> None:
    """Truncate table data.

    Args:
        table: Name of the table to truncate.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        None
    """
    raise NotImplementedError("truncate_table not implemented")


def rename_table(
    table: str, new_name: str, *args, namespace: Optional[str] = None, **kwargs
) -> None:
    """Rename an existing table.

    Args:
        table: Current name of the table.
        new_name: New name for the table.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not exist.
    """
    namespace = namespace or default_namespace()
    _get_storage(**kwargs).update_table(
        *args, table_name=table, new_table_name=new_name, namespace=namespace, **kwargs
    )


def table_exists(table: str, *args, namespace: Optional[str] = None, **kwargs) -> bool:
    """Check if a table exists in the catalog.

    Args:
        table: Name of the table to check.
        namespace: Optional namespace of the table. Uses default namespace if not specified.

    Returns:
        True if the table exists, False otherwise.
    """
    namespace = namespace or default_namespace()
    return _get_storage(**kwargs).table_exists(
        *args, table_name=table, namespace=namespace, **kwargs
    )


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """List a page of table namespaces.

    Args:
        catalog: Catalog properties instance.

    Returns:
        ListResult containing Namespace objects.
    """
    return _get_storage(**kwargs).list_namespaces(*args, **kwargs)


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """Get metadata for a specific table namespace.

    Args:
        namespace: Name of the namespace to retrieve.

    Returns:
        Namespace object if the namespace exists, None otherwise.
    """
    return _get_storage(**kwargs).get_namespace(*args, namespace=namespace, **kwargs)


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """Check if a namespace exists.

    Args:
        namespace: Name of the namespace to check.

    Returns:
        True if the namespace exists, False otherwise.
    """
    return _get_storage(**kwargs).namespace_exists(*args, namespace=namespace, **kwargs)


def create_namespace(
    namespace: str, *args, properties: Optional[NamespaceProperties] = None, **kwargs
) -> Namespace:
    """Create a new namespace.

    Args:
        namespace: Name of the namespace to create.
        properties: Optional properties for the namespace.

    Returns:
        Created Namespace object.

    Raises:
        NamespaceAlreadyExistsError: If the namespace already exists.
    """
    if namespace_exists(namespace, **kwargs):
        raise NamespaceAlreadyExistsError(f"Namespace {namespace} already exists")

    return _get_storage(**kwargs).create_namespace(
        *args, namespace=namespace, properties=properties, **kwargs
    )


def alter_namespace(
    namespace: str,
    *args,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    **kwargs,
) -> None:
    """Alter a namespace definition.

    Args:
        namespace: Name of the namespace to alter.
        properties: Optional new properties for the namespace.
        new_namespace: Optional new name for the namespace.

    Returns:
        None
    """
    _get_storage(**kwargs).update_namespace(
        namespace=namespace,
        properties=properties,
        new_namespace=new_namespace,
        *args,
        **kwargs,
    )


def drop_namespace(namespace: str, *args, purge: bool = False, **kwargs) -> None:
    """Drop a namespace and all of its tables from the catalog.

    Args:
        namespace: Name of the namespace to drop.
        purge: If True, permanently delete all tables in the namespace.
            If False, only remove from catalog.

    Returns:
        None

    TODO: Honor purge once garbage collection is implemented.
    """
    if purge:
        raise NotImplementedError("Purge flag is not currently supported.")

    _get_storage(**kwargs).delete_namespace(
        *args, namespace=namespace, purge=purge, **kwargs
    )


def default_namespace(*args, **kwargs) -> str:
    """Return the default namespace for the catalog.

    Returns:
        String name of the default namespace.
    """
    return DEFAULT_NAMESPACE  # table functions


def _validate_read_table_args(
    namespace: Optional[str] = None,
    table_type: Optional[TableType] = None,
    distributed_dataset_type: Optional[DistributedDatasetType] = None,
    merge_on_read: Optional[bool] = None,
    **kwargs,
):
    storage = _get_storage(**kwargs)
    if storage is None:
        raise ValueError(
            "Catalog not initialized. Did you miss calling "
            "initialize(ds=<deltacat_storage>)?"
        )

    if merge_on_read:
        raise ValueError("Merge on read not supported currently.")

    if table_type is not TableType.PYARROW:
        raise ValueError("Only PYARROW table type is supported as of now")

    if distributed_dataset_type is not DistributedDatasetType.DAFT:
        raise ValueError("Only DAFT dataset type is supported as of now")

    if namespace is None:
        raise ValueError(
            "namespace must be passed to uniquely identify a table in the catalog."
        )


def _get_latest_or_given_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> TableVersion:
    table_version_obj = None
    if table_version is None:
        table_version_obj = _get_storage(**kwargs).get_latest_table_version(
            namespace=namespace, table_name=table_name, *args, **kwargs
        )
        table_version = table_version_obj.table_version
    else:
        table_version_obj = _get_storage(**kwargs).get_table_version(
            namespace=namespace,
            table_name=table_name,
            table_version=table_version,
            *args,
            **kwargs,
        )

    return table_version_obj


def _get_deltas_from_partition_filter(
    partition_filter: Optional[List[Union[Partition, PartitionLocator]]] = None,
    stream_position_range_inclusive: Optional[Tuple[int, int]] = None,
    *args,
    **kwargs,
):

    result_deltas = []
    start_stream_position, end_stream_position = stream_position_range_inclusive or (
        None,
        None,
    )
    for partition_like in partition_filter:
        deltas = (
            _get_storage(**kwargs)
            .list_partition_deltas(
                partition_like=partition_like,
                ascending_order=True,
                include_manifest=True,
                start_stream_position=start_stream_position,
                last_stream_position=end_stream_position,
                *args,
                **kwargs,
            )
            .all_items()
        )

        for delta in deltas:
            if (
                start_stream_position is None
                or delta.stream_position >= start_stream_position
            ) and (
                end_stream_position is None
                or delta.stream_position <= end_stream_position
            ):
                if delta.type == DeltaType.DELETE:
                    raise ValueError("DELETE type deltas are not supported")
                result_deltas.append(delta)

    return result_deltas


def _get_storage(**kwargs):
    """
    Returns the implementation of `deltacat.storage.interface` to use with this catalog

    This is configured in the `CatalogProperties` stored during initialization and passed through `delegate.py`
    """
    properties: Optional[CatalogProperties] = kwargs.get("inner")
    if properties is not None and properties.storage is not None:
        return properties.storage
    else:
        return storage_impl



class RowFilter:
    ...


class ColumnFilter:
    ...


class PartitionFilter:
    ...


class Pushdown:
    """Represents pushdown predicates to be applied for DeltaCAT Tables"""

    row_filter: RowFilter
    column_filter: ColumnFilter
    partition_filter: PartitionFilter
    limit: int



from typing import Any


import deltacat as dc
from deltacat.catalog import Catalog


def copy(source, destination):
    src_parts = source.split("/")
    src_parts = [part for part in src_parts if part]
    dst_parts = destination.split("/")
    dst_parts = [part for part in dst_parts if part]
    if not dc.is_initialized():
        raise ValueError("Catalog not initialized.")
    if len(src_parts) != len(dst_parts) and len(src_parts) != len(dst_parts) + 1:
        # TODO(pdames): Better error message.
        raise ValueError(
            f"Cannot copy {source} to {destination}. "
            f"Source and destination must share the same type."
        )
    src_obj = get(source)
    if len(src_parts) == 1:
        # copy the given catalog
        raise NotImplementedError
    elif len(src_parts) == 2:
        # TODO(pdames): Make catalog specification optional if there is only
        #  one catalog (e.g., auto-retrieve src_parts[0]/dst_parts[0])
        # copy the given namespace
        src_namespace_name = src_parts[1]
        dst_catalog_name = dst_parts[0]
        dst_namespace_name = dst_parts[1] if len(dst_parts) >= 2 else src_namespace_name
        new_namespace = dc.create_namespace(
            namespace=dst_namespace_name,
            properties=src_obj.properties,
            catalog=dst_catalog_name,
        )
        return new_namespace
    elif len(src_parts) == 3:
        # copy the given table
        raise NotImplementedError
    elif len(src_parts) == 4:
        # copy the given table version
        raise NotImplementedError
    elif len(src_parts) == 5:
        # copy the given stream
        raise NotImplementedError
    elif len(src_parts) == 6:
        # copy the given partition
        raise NotImplementedError
    elif len(src_parts) == 7:
        # copy the given partition delta
        raise NotImplementedError
    raise ValueError(f"Invalid path: {src_parts}")


def concat(source, destination):
    raise NotImplementedError


def delete(source):
    raise NotImplementedError


def move(source, destination):
    raise NotImplementedError


def list(path):
    raise NotImplementedError


def get(path) -> Any:
    parts = path.split("/")
    parts = [part for part in parts if part]
    if not dc.is_initialized():
        # TODO(pdames): Re-initialize DeltaCAT with all catalogs from the
        #  last session.
        raise ValueError("Catalog not initialized.")
    if len(parts) == 1:
        # TODO(pdames): Save all catalogs registered from the last session on
        #  disk so that users don't need to re-initialize them every time.
        # get the given catalog
        catalog_name = parts[0]
        return dc.get_catalog(catalog_name)
    elif len(parts) == 2:
        # get the given namespace
        catalog_name = parts[0]
        namespace_name = parts[1]
        return dc.get_namespace(
            namespace=namespace_name,
            catalog=catalog_name,
        )
    elif len(parts) == 3:
        # get the given table
        raise NotImplementedError
    elif len(parts) == 4:
        # get the given table version
        raise NotImplementedError
    elif len(parts) == 5:
        # get the given stream
        raise NotImplementedError
    elif len(parts) == 6:
        # get the given partition
        raise NotImplementedError
    elif len(parts) == 7:
        # get the given partition delta
        raise NotImplementedError
    raise ValueError(f"Invalid path: {path}")


def put(path, *args, **kwargs) -> Any:
    parts = path.split("/")
    parts = [part for part in parts if part]
    if len(parts) == 1:
        # TODO(pdames): Save all catalogs registered from the last session on
        #  disk so that users don't need to re-initialize them every time.
        # register the given catalog
        catalog_name = parts[0]
        # Initialize default catalog using kwargs
        catalog = Catalog(**kwargs)
        return dc.put_catalog(catalog_name, catalog)
    elif len(parts) == 2:
        # register the given namespace
        catalog_name = parts[0]
        namespace_name = parts[1]
        if not dc.is_initialized():
            # TODO(pdames): Re-initialize DeltaCAT with all catalogs from the
            #  last session.
            raise ValueError("Catalog not initialized.")
        new_namespace = dc.create_namespace(
            namespace=namespace_name,
            catalog=catalog_name,
            *args,
            **kwargs,
        )
        return new_namespace
    elif len(parts) == 3:
        # register the given table
        raise NotImplementedError
    elif len(parts) == 4:
        # register the given table version
        raise NotImplementedError
    elif len(parts) == 5:
        # register the given stream
        raise NotImplementedError
    elif len(parts) == 6:
        # register the given partition
        raise NotImplementedError
    elif len(parts) == 7:
        # register the given partition delta
        raise NotImplementedError
    raise ValueError(f"Invalid path: {path}")


def exists(path):
    raise NotImplementedError


def query(path, expression):
    raise NotImplementedError


def tail(path):
    raise NotImplementedError


def head(path):
    raise NotImplementedError



from .schema.schema import Schema
from .schema.schema import Field
from .dataset import Dataset
from .schema.schema import Datatype

__all__ = [
    "Schema",
    "Field",
    "Dataset",
    "Datatype",
]



from typing import Optional, Dict
from deltacat.exceptions import RetryableError

AVERAGE_FILE_PATH_COLUMN_SIZE_BYTES = 80
AVERAGE_POS_COLUMN_SIZE_BYTES = 4
XXHASH_BYTE_PER_RECORD = 8
MEMORY_BUFFER_RATE = 1.2


def estimate_fixed_hash_columns(hash_value_size_bytes_per_record, total_record_count):
    return hash_value_size_bytes_per_record * total_record_count


def get_total_record_from_iceberg_files(iceberg_files_list):
    total_record_count = 0
    for iceberg_files in iceberg_files_list:
        total_record_count += sum(file.record_count for file in iceberg_files)
    return total_record_count


def estimate_iceberg_pos_delete_additional_columns(
    include_columns, num_of_record_count
):
    total_additional_columns_sizes = 0
    if "file_path" in include_columns:
        total_additional_columns_sizes += (
            AVERAGE_FILE_PATH_COLUMN_SIZE_BYTES * num_of_record_count
        )
    elif "pos" in include_columns:
        total_additional_columns_sizes += (
            AVERAGE_POS_COLUMN_SIZE_BYTES * num_of_record_count
        )
    return total_additional_columns_sizes


def estimate_convert_remote_option_resources(data_files, equality_delete_files):
    data_file_record_count = get_total_record_from_iceberg_files(data_files)
    equality_delete_record_count = get_total_record_from_iceberg_files(
        equality_delete_files
    )
    hash_column_sizes = estimate_fixed_hash_columns(
        XXHASH_BYTE_PER_RECORD, data_file_record_count + equality_delete_record_count
    )
    pos_delete_sizes = estimate_iceberg_pos_delete_additional_columns(
        ["file_path", "pos"], data_file_record_count + equality_delete_record_count
    )
    total_memory_required = hash_column_sizes + pos_delete_sizes
    return total_memory_required * MEMORY_BUFFER_RATE


def _get_task_options(
    memory: float,
    ray_custom_resources: Optional[Dict] = None,
    scheduling_strategy: str = "SPREAD",
) -> Dict:

    # NOTE: With DEFAULT scheduling strategy in Ray 2.20.0, autoscaler does
    # not spin up enough nodes fast and hence we see only approximately
    # 20 tasks get scheduled out of 100 tasks in queue. Hence, we use SPREAD
    # which is also ideal for merge and hash bucket tasks.
    # https://docs.ray.io/en/latest/ray-core/scheduling/index.html
    task_opts = {
        "memory": memory,
        "scheduling_strategy": scheduling_strategy,
    }

    if ray_custom_resources:
        task_opts["resources"] = ray_custom_resources

    task_opts["max_retries"] = 3

    # List of possible botocore exceptions are available at
    # https://github.com/boto/botocore/blob/develop/botocore/exceptions.py
    task_opts["retry_exceptions"] = [RetryableError]

    return task_opts


def convert_resource_options_provider(index, files_for_each_bucket):
    (
        data_files_list,
        equality_delete_files_list,
        position_delete_files_list,
    ) = files_for_each_bucket[1]
    memory_requirement = estimate_convert_remote_option_resources(
        data_files_list, equality_delete_files_list
    )
    return _get_task_options(memory=memory_requirement)



from __future__ import annotations
from typing import Dict


class ConvertInputFiles(Dict):
    @staticmethod
    def of(
        partition_value,
        all_data_files_for_dedupe=None,
        applicable_data_files=None,
        applicable_equality_delete_files=None,
        existing_position_delete_files=None,
    ) -> ConvertInputFiles:

        result = ConvertInputFiles()
        result["partition_value"] = partition_value
        result["all_data_files_for_dedupe"] = all_data_files_for_dedupe
        result["applicable_data_files"] = applicable_data_files
        result["applicable_equality_delete_files"] = applicable_equality_delete_files
        result["existing_position_delete_files"] = existing_position_delete_files
        return result

    @property
    def partition_value(self):
        return self["partition_value"]

    @property
    def all_data_files_for_dedupe(self):
        return self["all_data_files_for_dedupe"]

    @property
    def applicable_data_files(self):
        return self["applicable_data_files"]

    @property
    def applicable_equality_delete_files(self):
        return self["applicable_equality_delete_files"]

    @property
    def existing_position_delete_files(self):
        return self["existing_position_delete_files"]

    @partition_value.setter
    def partition_value(self, partition_value):
        self["partition_value"] = partition_value

    @all_data_files_for_dedupe.setter
    def all_data_files_for_dedupe(self, all_data_files_for_dedupe):
        self["all_data_files_for_dedupe"] = all_data_files_for_dedupe

    @applicable_data_files.setter
    def applicable_data_files(self, applicable_data_files):
        self["applicable_data_files"] = applicable_data_files

    @applicable_equality_delete_files.setter
    def applicable_equality_delete_files(self, applicable_equality_delete_files):
        self["applicable_equality_delete_files"] = applicable_equality_delete_files

    @existing_position_delete_files.setter
    def existing_position_delete_files(self, existing_position_delete_files):
        self["existing_position_delete_files"] = existing_position_delete_files
